# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Union, cast

import torch

import theseus
import theseus.constants

from .lie_group import LieGroup
from .point_types import Point3
from .so3 import SO3


class SE3(LieGroup):
    def __init__(
        self,
        x_y_z_quaternion: Optional[torch.Tensor] = None,
        tensor: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
        dtype: Optional[torch.dtype] = None,
        strict: bool = False,
    ):
        if x_y_z_quaternion is not None and tensor is not None:
            raise ValueError("Please provide only one of x_y_z_quaternion or tensor.")
        if x_y_z_quaternion is not None:
            dtype = x_y_z_quaternion.dtype
        super().__init__(tensor=tensor, name=name, dtype=dtype, strict=strict)
        if x_y_z_quaternion is not None:
            self.update_from_x_y_z_quaternion(x_y_z_quaternion=x_y_z_quaternion)

        self._resolve_eps()

    def _resolve_eps(self):
        self._NEAR_ZERO_EPS = theseus.constants._SE3_NEAR_ZERO_EPS[self.tensor.dtype]
        self._NEAR_PI_EPS = theseus.constants._SE3_NEAR_PI_EPS[self.tensor.dtype]

    @staticmethod
    def rand(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "SE3":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        ret = SE3()
        rotation = SO3.rand(
            size[0],
            generator=generator,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
        )
        translation = Point3.rand(
            size[0],
            generator=generator,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
        )
        ret.update_from_rot_and_trans(rotation=rotation, translation=translation)
        return ret

    @staticmethod
    def randn(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "SE3":
        if len(size) != 1:
            raise ValueError("The size should be 1D.")
        ret = SE3()
        rotation = SO3.randn(
            size[0],
            generator=generator,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
        )
        translation = Point3.randn(
            size[0],
            generator=generator,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
        )
        ret.update_from_rot_and_trans(rotation=rotation, translation=translation)
        return ret

    @staticmethod
    def _init_tensor() -> torch.Tensor:  # type: ignore
        return torch.eye(3, 4).view(1, 3, 4)

    def dof(self) -> int:
        return 6

    def __repr__(self) -> str:
        return f"SE3(tensor={self.tensor}, name={self.name})"

    def __str__(self) -> str:
        with torch.no_grad():
            return f"SE3(matrix={self.tensor}), name={self.name})"

    def _adjoint_impl(self) -> torch.Tensor:
        ret = torch.zeros(self.shape[0], 6, 6).to(dtype=self.dtype, device=self.device)
        ret[:, :3, :3] = self[:, :3, :3]
        ret[:, 3:, 3:] = self[:, :3, :3]
        ret[:, :3, 3:] = SO3.hat(self[:, :3, 3]) @ self[:, :3, :3]

        return ret

    def _project_impl(
        self, euclidean_grad: torch.Tensor, is_sparse: bool = False
    ) -> torch.Tensor:
        self._project_check(euclidean_grad, is_sparse)
        ret = torch.zeros(
            euclidean_grad.shape[:-2] + torch.Size([6]),
            dtype=self.dtype,
            device=self.device,
        )

        if is_sparse:
            temp = torch.einsum(
                "i...jk,i...jl->i...lk", euclidean_grad, self.tensor[:, :, :3]
            )
        else:
            temp = torch.einsum(
                "...jk,...ji->...ik", euclidean_grad, self.tensor[:, :, :3]
            )

        ret[..., :3] = temp[..., 3]
        ret[..., 3] = temp[..., 2, 1] - temp[..., 1, 2]
        ret[..., 4] = temp[..., 0, 2] - temp[..., 2, 0]
        ret[..., 5] = temp[..., 1, 0] - temp[..., 0, 1]

        return ret

    @staticmethod
    def _check_tensor_impl(tensor: torch.Tensor) -> bool:
        with torch.no_grad():
            if tensor.ndim != 3 or tensor.shape[1:] != (3, 4):
                raise ValueError("SE3 data tensors can only be 3x4 matrices.")

            return SO3._check_tensor_impl(tensor[:, :3, :3])

    @staticmethod
    def x_y_z_unit_quaternion_to_SE3(x_y_z_quaternion: torch.Tensor) -> "SE3":
        if x_y_z_quaternion.ndim == 1:
            x_y_z_quaternion = x_y_z_quaternion.unsqueeze(0)

        if x_y_z_quaternion.ndim != 2 and x_y_z_quaternion.shape[1] != 7:
            raise ValueError("x_y_z_quaternion can only be 7-D vectors.")

        ret = SE3()

        batch_size = x_y_z_quaternion.shape[0]
        ret.tensor = torch.empty(batch_size, 3, 4).to(
            device=x_y_z_quaternion.device, dtype=x_y_z_quaternion.dtype
        )
        ret[:, :, :3] = SO3.unit_quaternion_to_SO3(x_y_z_quaternion[:, 3:]).tensor
        ret[:, :, 3] = x_y_z_quaternion[:, :3]

        return ret

    @staticmethod
    def _hat_matrix_check(matrix: torch.Tensor):
        HAT_EPS = theseus.constants._SE3_HAT_EPS[matrix.dtype]

        if matrix.ndim != 3 or matrix.shape[1:] != (4, 4):
            raise ValueError("Hat matrices of SE3 can only be 4x4 matrices")

        if matrix[:, 3].abs().max().item() > HAT_EPS:
            raise ValueError("The last row of hat matrices of SE3 can only be zero.")

        if (
            matrix[:, :3, :3].transpose(1, 2) + matrix[:, :3, :3]
        ).abs().max().item() > HAT_EPS:
            raise ValueError(
                "The 3x3 top-left corner of hat matrices of SE3 can only be skew-symmetric."
            )

    @staticmethod
    def exp_map(
        tangent_vector: torch.Tensor, jacobians: Optional[List[torch.Tensor]] = None
    ) -> "SE3":
        if tangent_vector.ndim != 2 or tangent_vector.shape[1] != 6:
            raise ValueError("Tangent vectors of SE3 can only be 6D vectors.")

        ret = SE3(dtype=tangent_vector.dtype)

        tangent_vector_lin = tangent_vector[:, :3].view(-1, 3, 1)
        tangent_vector_ang = tangent_vector[:, 3:].view(-1, 3, 1)

        theta = torch.linalg.norm(tangent_vector_ang, dim=1).unsqueeze(1)
        theta2 = theta**2
        theta3 = theta**3

        near_zero = theta < theseus.constants._SE3_NEAR_ZERO_EPS[tangent_vector.dtype]
        non_zero = torch.ones(
            1, dtype=tangent_vector.dtype, device=tangent_vector.device
        )
        theta_nz = torch.where(near_zero, non_zero, theta)
        theta2_nz = torch.where(near_zero, non_zero, theta2)
        theta3_nz = torch.where(near_zero, non_zero, theta3)

        # Compute the rotation
        sine = theta.sin()
        cosine = torch.where(near_zero, 8 / (4 + theta2) - 1, theta.cos())
        sine_by_theta = torch.where(
            near_zero, 0.5 * cosine + 0.5, theta.sin() / theta_nz
        )
        one_minus_cosine_by_theta2 = torch.where(
            near_zero, 0.5 * sine_by_theta, (1 - cosine) / theta2_nz
        )
        ret.tensor = torch.zeros(tangent_vector.shape[0], 3, 4).to(
            dtype=tangent_vector.dtype, device=tangent_vector.device
        )
        ret.tensor[:, :3, :3] = (
            one_minus_cosine_by_theta2
            * tangent_vector_ang
            @ tangent_vector_ang.transpose(1, 2)
        )

        ret[:, 0, 0] += cosine.view(-1)
        ret[:, 1, 1] += cosine.view(-1)
        ret[:, 2, 2] += cosine.view(-1)
        temp = sine_by_theta.view(-1, 1) * tangent_vector_ang.view(-1, 3)
        ret[:, 0, 1] -= temp[:, 2]
        ret[:, 1, 0] += temp[:, 2]
        ret[:, 0, 2] += temp[:, 1]
        ret[:, 2, 0] -= temp[:, 1]
        ret[:, 1, 2] -= temp[:, 0]
        ret[:, 2, 1] += temp[:, 0]

        # Compute the translation
        sine_by_theta = torch.where(near_zero, 1 - theta2 / 6, sine_by_theta)
        one_minus_cosine_by_theta2 = torch.where(
            near_zero, 0.5 - theta2 / 24, one_minus_cosine_by_theta2
        )
        theta_minus_sine_by_theta3_t = torch.where(
            near_zero, 1.0 / 6 - theta2 / 120, (theta - sine) / theta3_nz
        )

        ret[:, :, 3:] = sine_by_theta * tangent_vector_lin
        ret[:, :, 3:] += one_minus_cosine_by_theta2 * torch.cross(
            tangent_vector_ang, tangent_vector_lin, dim=1
        )
        ret[:, :, 3:] += theta_minus_sine_by_theta3_t * (
            tangent_vector_ang
            @ (tangent_vector_ang.transpose(1, 2) @ tangent_vector_lin)
        )

        if jacobians is not None:
            SE3._check_jacobians_list(jacobians)
            theta3_nz = theta_nz * theta2_nz
            theta_minus_sine_by_theta3_rot = torch.where(
                near_zero, torch.zeros_like(theta), theta_minus_sine_by_theta3_t
            )
            jac = torch.zeros(
                tangent_vector.shape[0],
                6,
                6,
                dtype=tangent_vector.dtype,
                device=tangent_vector.device,
            )
            jac[:, :3, :3] = (
                theta_minus_sine_by_theta3_rot
                * tangent_vector_ang.view(-1, 3, 1)
                @ tangent_vector_ang.view(-1, 1, 3)
            )
            diag_jac = jac.diagonal(dim1=1, dim2=2)
            diag_jac += sine_by_theta.view(-1, 1)

            jac_temp_rot = one_minus_cosine_by_theta2.view(
                -1, 1
            ) * tangent_vector_ang.view(-1, 3)

            jac[:, 0, 1] += jac_temp_rot[:, 2]
            jac[:, 1, 0] -= jac_temp_rot[:, 2]
            jac[:, 0, 2] -= jac_temp_rot[:, 1]
            jac[:, 2, 0] += jac_temp_rot[:, 1]
            jac[:, 1, 2] += jac_temp_rot[:, 0]
            jac[:, 2, 1] -= jac_temp_rot[:, 0]

            jac[:, 3:, 3:] = jac[:, :3, :3]

            minus_one_by_twelve = torch.tensor(
                -1 / 12.0,
                dtype=sine_by_theta.dtype,
                device=sine_by_theta.device,
            )
            d_one_minus_cosine_by_theta2 = torch.where(
                near_zero,
                minus_one_by_twelve,
                (sine_by_theta - 2 * one_minus_cosine_by_theta2) / theta2_nz,
            )
            minus_one_by_sixty = torch.tensor(
                -1 / 60.0,
                dtype=one_minus_cosine_by_theta2.dtype,
                device=one_minus_cosine_by_theta2.device,
            )
            d_theta_minus_sine_by_theta3 = torch.where(
                near_zero,
                minus_one_by_sixty,
                (one_minus_cosine_by_theta2 - 3 * theta_minus_sine_by_theta3_t)
                / theta2_nz,
            )

            w = tangent_vector[:, 3:]
            v = tangent_vector[:, :3]
            wv = w.cross(v, dim=1)
            wwv = w.cross(wv, dim=1)
            sw = theta_minus_sine_by_theta3_t.view(-1, 1) * w

            jac_temp_t = (
                d_one_minus_cosine_by_theta2.view(-1, 1) * wv
                + d_theta_minus_sine_by_theta3.view(-1, 1) * wwv
            ).view(-1, 3, 1) @ w.view(-1, 1, 3)
            jac_temp_t -= v.view(-1, 3, 1) @ sw.view(-1, 1, 3)
            jac_temp_v = (
                -one_minus_cosine_by_theta2.view(-1, 1) * v
                - theta_minus_sine_by_theta3_t.view(-1, 1) * wv
            )
            jac_temp_t += SO3.hat(jac_temp_v)
            diag_jac_t = torch.diagonal(jac_temp_t, dim1=1, dim2=2)
            diag_jac_t += (sw.view(-1, 1, 3) @ v.view(-1, 3, 1)).view(-1, 1)

            jac[:, :3, 3:] = ret[:, :, :3].transpose(1, 2) @ jac_temp_t

            jacobians.append(jac)

        return ret

    @staticmethod
    def normalize(tensor: torch.Tensor) -> torch.Tensor:
        if tensor.ndim != 3 or tensor.shape[1:] != (3, 4):
            raise ValueError("SE3 data tensors can only be 3x4 matrices.")

        return torch.cat([SO3.normalize(tensor[:, :, :3]), tensor[:, :, 3:]], dim=-1)

    def _log_map_impl(
        self, jacobians: Optional[List[torch.Tensor]] = None
    ) -> torch.Tensor:

        sine_axis = torch.zeros(self.shape[0], 3, dtype=self.dtype, device=self.device)
        sine_axis[:, 0] = 0.5 * (self[:, 2, 1] - self[:, 1, 2])
        sine_axis[:, 1] = 0.5 * (self[:, 0, 2] - self[:, 2, 0])
        sine_axis[:, 2] = 0.5 * (self[:, 1, 0] - self[:, 0, 1])
        cosine = 0.5 * (self[:, 0, 0] + self[:, 1, 1] + self[:, 2, 2] - 1)
        sine = sine_axis.norm(dim=1)
        theta = torch.atan2(sine, cosine)
        theta2 = theta**2
        non_zero = torch.ones(1, dtype=self.dtype, device=self.device)

        near_zero = theta < self._NEAR_ZERO_EPS

        # Compute the rotation
        not_near_pi = 1 + cosine > self._NEAR_PI_EPS
        # theta is not near pi
        near_zero_not_near_pi = near_zero[not_near_pi]
        # Compute the approximation of theta / sin(theta) when theta is near to 0
        sine_nz = torch.where(near_zero_not_near_pi, non_zero, sine[not_near_pi])
        scale = torch.where(
            near_zero_not_near_pi,
            1 + sine[not_near_pi] ** 2 / 6,
            theta[not_near_pi] / sine_nz,
        )
        ret_ang = torch.zeros_like(sine_axis)
        ret_ang[not_near_pi] = sine_axis[not_near_pi] * scale.view(-1, 1)

        # theta is near pi
        near_pi = ~not_near_pi
        ddiag = torch.diagonal(self[near_pi], dim1=1, dim2=2)
        # Find the index of major coloumns and diagonals
        major = torch.logical_and(
            ddiag[:, 1] > ddiag[:, 0], ddiag[:, 1] > ddiag[:, 2]
        ) + 2 * torch.logical_and(ddiag[:, 2] > ddiag[:, 0], ddiag[:, 2] > ddiag[:, 1])
        sel_rows = 0.5 * (self[near_pi, major, :3] + self[near_pi, :3, major])
        aux = torch.ones(sel_rows.shape[0], dtype=torch.bool)
        sel_rows[aux, major] -= cosine[near_pi]
        axis = sel_rows / sel_rows.norm(dim=1, keepdim=True)
        sign_tmp = sine_axis[near_pi, major].sign()
        sign = torch.where(sign_tmp != 0, sign_tmp, torch.ones_like(sign_tmp))
        ret_ang[near_pi] = axis * (theta[near_pi] * sign).view(-1, 1)

        # Compute the translation
        sine_theta = sine * theta
        two_cosine_minus_two = 2 * cosine - 2
        two_cosine_minus_two_nz = torch.where(near_zero, non_zero, two_cosine_minus_two)

        theta2_nz = torch.where(near_zero, non_zero, theta2)

        a = torch.where(
            near_zero, 1 - theta2 / 12, -sine_theta / two_cosine_minus_two_nz
        )
        b = torch.where(
            near_zero,
            1.0 / 12 + theta2 / 720,
            (sine_theta + two_cosine_minus_two) / (theta2_nz * two_cosine_minus_two_nz),
        )

        translation = self[:, :, 3].view(-1, 3, 1)
        ret_lin = a.view(-1, 1) * self[:, :, 3]
        ret_lin -= 0.5 * torch.cross(ret_ang, self[:, :, 3], dim=1)
        ret_ang_ext = ret_ang.view(-1, 3, 1)
        ret_lin += b.view(-1, 1) * (
            ret_ang_ext @ (ret_ang_ext.transpose(1, 2) @ translation)
        ).view(-1, 3)

        if jacobians is not None:
            SE3._check_jacobians_list(jacobians)
            jac = torch.zeros(self.shape[0], 6, 6, dtype=self.dtype, device=self.device)

            b_ret_ang = b.view(-1, 1) * ret_ang
            jac[:, :3, :3] = b_ret_ang.view(-1, 3, 1) * ret_ang.view(-1, 1, 3)

            half_ret_ang = 0.5 * ret_ang
            jac[:, 0, 1] -= half_ret_ang[:, 2]
            jac[:, 1, 0] += half_ret_ang[:, 2]
            jac[:, 0, 2] += half_ret_ang[:, 1]
            jac[:, 2, 0] -= half_ret_ang[:, 1]
            jac[:, 1, 2] -= half_ret_ang[:, 0]
            jac[:, 2, 1] += half_ret_ang[:, 0]

            diag_jac_rot = torch.diagonal(jac[:, :3, :3], dim1=1, dim2=2)
            diag_jac_rot += a.view(-1, 1)

            jac[:, 3:, 3:] = jac[:, :3, :3]

            theta_nz = torch.where(near_zero, non_zero, theta)
            theta4_nz = theta2_nz**2
            c = torch.where(
                near_zero,
                -1 / 360.0 - theta2 / 7560.0,
                -(2 * two_cosine_minus_two + theta * sine + theta2)
                / (theta4_nz * two_cosine_minus_two_nz),
            )
            d = torch.where(
                near_zero,
                -1 / 6.0 - theta2 / 180.0,
                (theta - sine) / (theta_nz * two_cosine_minus_two_nz),
            )
            e = (ret_ang.view(-1, 1, 3) @ ret_lin.view(-1, 3, 1)).view(-1)

            ce_ret_ang = (c * e).view(-1, 1) * ret_ang
            jac[:, :3, 3:] = ce_ret_ang.view(-1, 3, 1) * ret_ang.view(-1, 1, 3)
            jac[:, :3, 3:] += b_ret_ang.view(-1, 3, 1) * ret_lin.view(
                -1, 1, 3
            ) + ret_lin.view(-1, 3, 1) * b_ret_ang.view(-1, 1, 3)
            diag_jac_t = torch.diagonal(jac[:, :3, 3:], dim1=1, dim2=2)
            diag_jac_t += (e * d).view(-1, 1)

            half_ret_lin = 0.5 * ret_lin
            jac[:, 0, 4] -= half_ret_lin[:, 2]
            jac[:, 1, 3] += half_ret_lin[:, 2]
            jac[:, 0, 5] += half_ret_lin[:, 1]
            jac[:, 2, 3] -= half_ret_lin[:, 1]
            jac[:, 1, 5] -= half_ret_lin[:, 0]
            jac[:, 2, 4] += half_ret_lin[:, 0]

            jacobians.append(jac)

        return torch.cat([ret_lin, ret_ang], dim=1)

    def _compose_impl(self, se3_2: LieGroup) -> "SE3":
        se3_2 = cast(SE3, se3_2)
        batch_size = max(self.shape[0], se3_2.shape[0])
        ret = SE3()
        ret.tensor = torch.zeros(batch_size, 3, 4, dtype=self.dtype, device=self.device)
        ret[:, :, :3] = self[:, :, :3] @ se3_2[:, :, :3]
        ret[:, :, 3] = self[:, :, 3]
        ret[:, :, 3:] += self[:, :, :3] @ se3_2[:, :, 3:]

        return ret

    def _inverse_impl(self, get_jacobian: bool = False) -> "SE3":
        ret = torch.zeros(self.shape[0], 3, 4).to(dtype=self.dtype, device=self.device)
        rotT = self.tensor[:, :3, :3].transpose(1, 2)
        ret[:, :, :3] = rotT
        ret[:, :, 3] = -(rotT @ self.tensor[:, :3, 3].unsqueeze(2)).view(-1, 3)
        # if self.tensor is a valid SE3, so is the inverse
        return SE3(tensor=ret, strict=False)

    def to_matrix(self) -> torch.Tensor:
        ret = torch.zeros(self.shape[0], 4, 4).to(dtype=self.dtype, device=self.device)
        ret[:, :3] = self.tensor
        ret[:, 3, 3] = 1
        return ret

    def update_from_x_y_z_quaternion(self, x_y_z_quaternion: torch.Tensor):
        self.update(SE3.x_y_z_unit_quaternion_to_SE3(x_y_z_quaternion))

    def update_from_rot_and_trans(self, rotation: SO3, translation: Point3):
        if rotation.shape[0] != translation.shape[0]:
            raise ValueError("rotation and translation must have the same batch size.")

        if rotation.dtype != translation.dtype:
            raise ValueError("rotation and translation must be of the same type.")

        if rotation.device != translation.device:
            raise ValueError("rotation and translation must be on the same device.")

        self.tensor = torch.cat(
            (rotation.tensor, translation.tensor.unsqueeze(2)), dim=2
        )

    @staticmethod
    def hat(tangent_vector: torch.Tensor) -> torch.Tensor:
        _check = tangent_vector.ndim == 2 and tangent_vector.shape[1] == 6
        if not _check:
            raise ValueError("Invalid vee matrix for SE3.")
        matrix = torch.zeros(tangent_vector.shape[0], 4, 4).to(
            dtype=tangent_vector.dtype, device=tangent_vector.device
        )
        matrix[:, :3, :3] = SO3.hat(tangent_vector[:, 3:])
        matrix[:, :3, 3] = tangent_vector[:, :3]

        return matrix

    @staticmethod
    def vee(matrix: torch.Tensor) -> torch.Tensor:
        SE3._hat_matrix_check(matrix)
        return torch.cat((matrix[:, :3, 3], SO3.vee(matrix[:, :3, :3])), dim=1)

    def _copy_impl(self, new_name: Optional[str] = None) -> "SE3":
        # if self.tensor is a valid SE3, so is the copy
        return SE3(tensor=self.tensor.clone(), name=new_name, strict=False)

    # only added to avoid casting downstream
    def copy(self, new_name: Optional[str] = None) -> "SE3":
        return cast(SE3, super().copy(new_name=new_name))

    def _transform_shape_check(self, point: Union[Point3, torch.Tensor]):
        err_msg = (
            f"SE3 can only transform vectors of shape [{self.shape[0]}, 3] or [1, 3], "
            f"but the input has shape {point.shape}."
        )

        if isinstance(point, torch.Tensor):
            if not point.ndim == 2 or point.shape[1] != 3:
                raise ValueError(err_msg)
        elif point.dof() != 3:
            raise ValueError(err_msg)
        if (
            point.shape[0] != self.shape[0]
            and point.shape[0] != 1
            and self.shape[0] != 1
        ):
            raise ValueError(err_msg)

    def transform_from(
        self,
        point: Union[Point3, torch.Tensor],
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> Point3:
        self._transform_shape_check(point)
        batch_size = max(self.shape[0], point.shape[0])
        if isinstance(point, torch.Tensor):
            p = point.view(-1, 3, 1)
        else:
            p = point.tensor.view(-1, 3, 1)

        ret = Point3(tensor=(self[:, :, :3] @ p).view(-1, 3))
        ret.tensor += self[:, :, 3]

        if jacobians is not None:
            self._check_jacobians_list(jacobians)
            # Right jacobians for SE3 are computed
            Jg = torch.zeros(batch_size, 3, 6, dtype=self.dtype, device=self.device)
            Jg[:, :, :3] = self[:, :, :3]
            Jg[:, :, 3:] = -self[:, :, :3] @ SO3.hat(p)
            # Jacobians for point
            Jpnt = Jg[:, :, :3]

            jacobians.extend([Jg, Jpnt])

        return ret

    def transform_to(
        self,
        point: Union[Point3, torch.Tensor],
        jacobians: Optional[List[torch.Tensor]] = None,
    ) -> Point3:
        self._transform_shape_check(point)
        batch_size = max(self.shape[0], point.shape[0])
        if isinstance(point, torch.Tensor):
            p = point.view(-1, 3, 1)
        else:
            p = point.tensor.view(-1, 3, 1)

        temp = p - self[:, :, 3:]
        ret = Point3(tensor=(self[:, :, :3].transpose(1, 2) @ temp).view(-1, 3))

        if jacobians is not None:
            self._check_jacobians_list(jacobians)
            # Right jacobians for SE3 are computed
            Jg = torch.zeros(batch_size, 3, 6, dtype=self.dtype, device=self.device)
            Jg[:, 0, 0] = -1
            Jg[:, 1, 1] = -1
            Jg[:, 2, 2] = -1
            Jg[:, 0, 4] = -ret[:, 2]
            Jg[:, 1, 3] = ret[:, 2]
            Jg[:, 0, 5] = ret[:, 1]
            Jg[:, 2, 3] = -ret[:, 1]
            Jg[:, 1, 5] = -ret[:, 0]
            Jg[:, 2, 4] = ret[:, 0]
            # Jacobians for point
            Jpnt = self[:, :, :3].transpose(1, 2).expand(batch_size, 3, 3)

            jacobians.extend([Jg, Jpnt])

        return ret

    # calls to() on the internal tensors
    def to(self, *args, **kwargs):
        super().to(*args, **kwargs)
        self._resolve_eps()


rand_se3 = SE3.rand
randn_se3 = SE3.randn



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from dataclasses import dataclass
from typing import Dict, Optional

import numpy as np
import torch

from theseus.core import Objective, Vectorize


# All info information is batched
@dataclass
class OptimizerInfo:
    best_solution: Optional[Dict[str, torch.Tensor]]
    # status for each element
    status: np.ndarray


class Optimizer(abc.ABC):
    def __init__(self, objective: Objective, *args, vectorize: bool = True, **kwargs):
        self.objective = objective
        if vectorize:
            Vectorize(
                self.objective, empty_cuda_cache=kwargs.get("empty_cuda_cache", False)
            )
        self._objectives_version = objective.current_version

    @abc.abstractmethod
    def _optimize_impl(self, **kwargs) -> OptimizerInfo:
        pass

    def optimize(self, **kwargs) -> OptimizerInfo:
        if self._objectives_version != self.objective.current_version:
            raise RuntimeError(
                "The objective was modified after optimizer construction, which is "
                "currently not supported."
            )
        return self._optimize_impl(**kwargs)



import torch


# Obtained from https://github.com/pytorch/pytorch/issues/34704#issuecomment-878940122
def grid_sample(image, optical):
    """
    Custom implementation for torch.nn.functional.grid_sample() to avoid this warning:
    > "RuntimeError: derivative for aten::grid_sampler_2d_backward is not implemented"
    """
    N, C, IH, IW = image.shape
    _, H, W, _ = optical.shape
    ix = optical[..., 0]
    iy = optical[..., 1]
    ix = ((ix + 1) / 2) * (IW - 1)
    iy = ((iy + 1) / 2) * (IH - 1)
    with torch.no_grad():
        ix_nw = torch.floor(ix)
        iy_nw = torch.floor(iy)
        ix_ne = ix_nw + 1
        iy_ne = iy_nw
        ix_sw = ix_nw
        iy_sw = iy_nw + 1
        ix_se = ix_nw + 1
        iy_se = iy_nw + 1
    nw = (ix_se - ix) * (iy_se - iy)
    ne = (ix - ix_sw) * (iy_sw - iy)
    sw = (ix_ne - ix) * (iy - iy_ne)
    se = (ix - ix_nw) * (iy - iy_nw)
    with torch.no_grad():
        ix_nw = torch.clamp(ix_nw, 0, IW - 1)
        iy_nw = torch.clamp(iy_nw, 0, IH - 1)
        ix_ne = torch.clamp(ix_ne, 0, IW - 1)
        iy_ne = torch.clamp(iy_ne, 0, IH - 1)
        ix_sw = torch.clamp(ix_sw, 0, IW - 1)
        iy_sw = torch.clamp(iy_sw, 0, IH - 1)
        ix_se = torch.clamp(ix_se, 0, IW - 1)
        iy_se = torch.clamp(iy_se, 0, IH - 1)
    image = image.view(N, C, IH * IW)
    nw_val = torch.gather(
        image, 2, (iy_nw * IW + ix_nw).long().view(N, 1, H * W).repeat(1, C, 1)
    )
    ne_val = torch.gather(
        image, 2, (iy_ne * IW + ix_ne).long().view(N, 1, H * W).repeat(1, C, 1)
    )
    sw_val = torch.gather(
        image, 2, (iy_sw * IW + ix_sw).long().view(N, 1, H * W).repeat(1, C, 1)
    )
    se_val = torch.gather(
        image, 2, (iy_se * IW + ix_se).long().view(N, 1, H * W).repeat(1, C, 1)
    )
    out_val = (
        nw_val.view(N, C, H, W) * nw.view(N, 1, H, W)
        + ne_val.view(N, C, H, W) * ne.view(N, 1, H, W)
        + sw_val.view(N, C, H, W) * sw.view(N, 1, H, W)
        + se_val.view(N, C, H, W) * se.view(N, 1, H, W)
    )
    return out_val



from typing import Dict, List, Optional, Tuple

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import torch

# ----------------------------------------------------------------------------------- #
# ------------------------------------ Data loading --------------------------------- #
# ----------------------------------------------------------------------------------- #


class TactilePushingDataset:
    def __init__(
        self,
        data_fname: str,
        sdf_fname: str,
        episode_length: int,
        batch_size: int,
        max_episodes: int,
        max_steps: int,
        device: torch.device,
        split_episodes: bool = False,
        data_mode: str = "all",
        val_ratio: float = 0.1,
        seed: int = 1234567,
    ):
        assert data_mode in ["all", "train", "val"]

        batch_size = min(batch_size, max_episodes)
        data = TactilePushingDataset._load_dataset_from_file(
            data_fname, episode_length, max_episodes, device, split_episodes
        )
        (
            self.sdf_data_tensor,
            self.sdf_cell_size,
            self.sdf_origin,
        ) = TactilePushingDataset._load_tactile_sdf_from_file(sdf_fname, device)

        # keys: img_feats, eff_poses, obj_poses, contact_episode, contact_flag
        num_episodes = data["obj_poses"].shape[0]
        if data_mode == "all":
            idx = np.arange(num_episodes)
        else:
            rng = np.random.default_rng(seed)
            order = rng.permutation(num_episodes)
            stop = max(int(np.ceil(num_episodes * val_ratio)), 2)
            idx = order[:stop] if data_mode == "val" else order[stop:]

        self.img_feats = data["img_feats"][idx]
        self.eff_poses = data["eff_poses"][idx]
        self.obj_poses = data["obj_poses"][idx]
        self.contact_episode = data["contact_episode"][idx]
        self.contact_flag = data["contact_flag"][idx]
        # Check sizes of the attributes assigned above
        self.dataset_size: int = -1
        for key in data:
            if self.dataset_size == -1:
                self.dataset_size = getattr(self, key).shape[0]
            else:
                assert self.dataset_size == getattr(self, key).shape[0]
        print(f"Dataset for mode '{data_mode}' has size {self.dataset_size}.")

        # obj_poses is shape (num_episodes, episode_length, 3)
        self.time_steps = np.minimum(max_steps, self.obj_poses.shape[1])
        self.batch_size = batch_size
        self.num_batches = (self.dataset_size - 1) // self.batch_size + 1

    @staticmethod
    def _load_dataset_from_file(
        filename: str,
        episode_length: int,
        max_episodes: int,
        device: torch.device,
        split_episodes: bool = False,
    ) -> Dict[str, torch.Tensor]:
        # Load all episode data
        with open(filename) as f:
            import json

            data_from_file = json.load(f)

        dataset_all = {}
        dataset_all["obj_poses"] = torch.tensor(
            data_from_file["obj_poses_2d"], device=device
        )
        dataset_all["eff_poses"] = torch.tensor(
            data_from_file["ee_poses_2d"], device=device
        )
        dataset_all["img_feats"] = torch.tensor(
            data_from_file["img_feats"], device=device
        )
        dataset_all["contact_episode"] = torch.tensor(
            data_from_file["contact_episode"], device=device
        )
        dataset_all["contact_flag"] = torch.tensor(
            data_from_file["contact_flag"], device=device
        )

        # Read all episodes and filter those with length less than desired
        episode_indices = [
            idx.item() for idx in dataset_all["contact_episode"].unique()
        ]
        data: Dict[str, List[torch.Tensor]] = dict(
            [(k, []) for k in dataset_all.keys()]
        )

        for episode in episode_indices:
            if len(data["obj_poses"]) >= max_episodes:
                break
            ds_idxs = torch.nonzero(dataset_all["contact_episode"] == episode).squeeze()
            if len(ds_idxs) < episode_length:
                continue
            for key, val in dataset_all.items():
                if split_episodes:
                    tensors = TactilePushingDataset._get_tensor_splits(
                        val[ds_idxs], episode_length
                    )
                else:
                    ds_idxs = ds_idxs[:episode_length]
                    tensors = [val[ds_idxs]]
                for tensor in tensors:
                    data[key].append(tensor)

        # Stack all episode data into single tensors
        data_tensors = {}
        for key in data:
            data_tensors[key] = torch.stack(data[key])
        num_ep_read, len_ep_read = data_tensors[key].shape[:2]
        print(f"Read {num_ep_read} episodes of length {len_ep_read}.")
        return data_tensors

    @staticmethod
    def _get_tensor_splits(
        tensor: torch.Tensor, episode_length: int
    ) -> List[torch.Tensor]:
        squeeze = False
        if tensor.ndim == 1:
            squeeze = True
            tensor = tensor.view(-1, 1)
        length, dof = tensor.shape
        num_splits = length // episode_length
        mod_tensor = tensor[: num_splits * episode_length]
        reshaped_tensor = mod_tensor.view(num_splits, -1, dof)
        if squeeze:
            reshaped_tensor = reshaped_tensor.squeeze(2)
        return [t for t in reshaped_tensor]

    @staticmethod
    def _load_tactile_sdf_from_file(
        filename: str, device: torch.device
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

        with open(filename) as f:
            import json

            sdf_from_file = json.load(f)
        sdf_data_vec = sdf_from_file["grid_data"]

        sdf_data_mat = np.zeros(
            (sdf_from_file["grid_size_y"], sdf_from_file["grid_size_x"])
        )
        for i in range(sdf_data_mat.shape[0]):
            for j in range(sdf_data_mat.shape[1]):
                sdf_data_mat[i, j] = sdf_data_vec[i][j]

        sdf_data_tensor = torch.tensor(sdf_data_mat, device=device).unsqueeze(0)
        cell_size = torch.tensor([sdf_from_file["grid_res"]], device=device).unsqueeze(
            0
        )
        origin = torch.tensor(
            [sdf_from_file["grid_origin_x"], sdf_from_file["grid_origin_y"]],
            device=device,
        ).unsqueeze(0)

        return sdf_data_tensor, cell_size, origin

    def get_batch(self, batch_idx: int) -> Dict[str, torch.Tensor]:
        assert batch_idx < self.num_batches
        start = batch_idx * self.batch_size
        end = min(start + self.batch_size, self.dataset_size)
        batch = {}
        batch["img_feats"] = self.img_feats[start:end, : self.time_steps]
        batch["eff_poses"] = self.eff_poses[start:end, : self.time_steps]
        batch["obj_poses"] = self.obj_poses[start:end, : self.time_steps]
        batch["obj_poses_gt"] = self.obj_poses[start:end, : self.time_steps, :].clone()
        batch["eff_poses_gt"] = self.eff_poses[start:end, : self.time_steps, :].clone()
        batch["obj_start_pose"] = self.obj_poses[start:end, 0]
        for i in range(self.time_steps):
            batch[f"motion_capture_{i}"] = self.eff_poses[start:end, i]
        return batch


# ----------------------------------------------------------------------------------- #
# ---------------------------------- Visualization ---------------------------------- #
# ----------------------------------------------------------------------------------- #
def _draw_tactile_effector(poses: np.ndarray, label: str = "groundtruth"):

    linestyle = "-"
    color = "tab:orange"
    if label == "groundtruth":
        linestyle = "--"
        color = "tab:gray"

    # plot contact point and normal
    plt.plot(poses[-1][0], poses[-1][1], "k*", linestyle=linestyle)
    ori = poses[-1][2]
    (dx, dy) = (0.03 * -np.sin(ori), 0.03 * np.cos(ori))
    plt.arrow(
        poses[-1][0],
        poses[-1][1],
        dx,
        dy,
        linewidth=2,
        head_width=0.001,
        color=color,
        head_length=0.01,
        fc=color,
        ec=color,
    )

    eff_radius = 0.0075
    circle = mpatches.Circle(
        (poses[-1][0], poses[-1][1]), color=color, radius=eff_radius
    )
    plt.gca().add_patch(circle)


def _draw_tactile_object(
    poses: np.ndarray, rect_len_x: float, rect_len_y: float, label: str = "optimizer"
):

    linestyle = "-"
    color = "tab:orange"
    if label == "groundtruth":
        linestyle = "--"
        color = "tab:gray"

    plt.plot(
        poses[:, 0],
        poses[:, 1],
        color=color,
        linestyle=linestyle,
        label=label,
        linewidth=2,
        alpha=0.9,
    )

    # shape: rect
    yaw = poses[-1][2]
    R = np.array([[np.cos(yaw), -np.sin(yaw)], [np.sin(yaw), np.cos(yaw)]])
    offset = np.matmul(R, np.array([[0.5 * rect_len_x], [0.5 * rect_len_y]]))
    xb, yb = poses[-1][0] - offset[0], poses[-1][1] - offset[1]
    rect = mpatches.Rectangle(
        (xb, yb),
        rect_len_x,
        rect_len_y,
        angle=(np.rad2deg(yaw)),
        facecolor="None",
        edgecolor=color,
        linestyle=linestyle,
        linewidth=2,
    )
    plt.gca().add_patch(rect)


def visualize_tactile_push2d(
    obj_poses: torch.Tensor,
    eff_poses: torch.Tensor,
    obj_poses_gt: torch.Tensor,
    eff_poses_gt: torch.Tensor,
    rect_len_x: float,
    rect_len_y: float,
    save_fname: Optional[str] = None,
):

    plt.cla()
    plt.xlim((-2, 2))
    plt.ylim((-2, 2))
    plt.gca().axis("equal")
    plt.axis("off")

    _draw_tactile_object(
        obj_poses.cpu().detach().numpy(), rect_len_x, rect_len_y, label="optimizer"
    )
    _draw_tactile_effector(eff_poses.cpu().detach().numpy(), label="optimizer")
    _draw_tactile_object(
        obj_poses_gt.cpu().detach().numpy(), rect_len_x, rect_len_y, label="groundtruth"
    )
    _draw_tactile_effector(eff_poses_gt.cpu().detach().numpy(), label="groundtruth")

    plt.show()
    plt.pause(1e-9)

    if save_fname is not None:
        plt.savefig(save_fname)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th

from .common import MockVar


def test_variable_init():
    all_ids = []
    for i in range(100):
        if np.random.random() < 0.5:
            name = f"name_{i}"
        else:
            name = None
        data = torch.rand(1, 1)
        t = th.Variable(data, name=name)
        all_ids.append(t._id)
        if name is not None:
            assert name == t.name
        else:
            assert t.name == f"Variable__{t._id}"
        assert t.tensor.allclose(data)

    assert len(set(all_ids)) == len(all_ids)


def test_properties():
    for i in range(100):
        length = np.random.randint(2, 5)
        batch_size = np.random.randint(1, 100)
        dtype = torch.float if np.random.random() < 0.5 else torch.long
        data = torch.ones(batch_size, length, dtype=dtype)
        t = th.Variable(data)
        assert t.shape == data.shape
        assert t.ndim == data.ndim
        assert t.dtype == dtype


def test_update():
    for _ in range(10):
        for length in range(1, 10):
            var = MockVar(length)
            batch_size = np.random.randint(1, 10)
            # check update from torch tensor
            new_data_good = torch.rand(batch_size, length)
            var.update(new_data_good)
            assert var.tensor is new_data_good
            # check update from variable
            new_data_good_wrapped = torch.rand(batch_size, length)
            another_var = MockVar(length, tensor=new_data_good_wrapped)
            var.update(another_var)
            # check raises error on shape
            new_data_bad = torch.rand(batch_size, length + 1)
            with pytest.raises(ValueError):
                var.update(new_data_bad)
            # check raises error on dtype
            new_data_bad = torch.rand(batch_size, length + 1).double()
            with pytest.raises(ValueError):
                var.update(new_data_bad)
            # check batch indices to ignore
            how_many = np.random.randint(1, batch_size + 1)
            ignore_indices = np.random.choice(batch_size, size=how_many)
            ignore_mask = torch.zeros(batch_size).bool()
            ignore_mask[ignore_indices] = 1
            old_data = var.tensor.clone()
            new_data_some_ignored = torch.randn(batch_size, length)
            if ignore_indices[0] % 2 == 0:  # randomly wrap into a variable to also test
                new_data_some_ignored = MockVar(length, new_data_some_ignored)
            var.update(new_data_some_ignored, batch_ignore_mask=ignore_mask)
            for i in range(batch_size):
                if ignore_mask[i] == 1:
                    assert torch.allclose(var[i], old_data[i])
                else:
                    if isinstance(new_data_some_ignored, th.Variable):
                        new_data_some_ignored = new_data_some_ignored.tensor
                    assert torch.allclose(var[i], new_data_some_ignored[i])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Optional, Type

import torch

# rename Cholesky `Factor` to CholeskyDecomposition to
# prevent confusion with the factors of the factor graph,
# when using the probabilistic naming convention
from sksparse.cholmod import Factor as CholeskyDecomposition
from sksparse.cholmod import analyze_AAt

from theseus.core import Objective
from theseus.optimizer import Linearization, SparseLinearization
from theseus.optimizer.autograd import CholmodSolveFunction

from .linear_solver import LinearSolver


class CholmodSparseSolver(LinearSolver):
    def __init__(
        self,
        objective: Objective,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        damping: float = 1e-6,
        **kwargs,
    ):
        linearization_cls = linearization_cls or SparseLinearization
        if not linearization_cls == SparseLinearization:
            raise RuntimeError(
                "CholmodSparseSolver only works with theseus.optimizer.SparseLinearization,"
                + f" got {type(self.linearization)}"
            )

        super().__init__(objective, linearization_cls, linearization_kwargs, **kwargs)
        self.linearization: SparseLinearization = self.linearization

        # symbolic decomposition depending on the sparse structure, done with mock data
        self._symbolic_cholesky_decomposition: CholeskyDecomposition = analyze_AAt(
            self.linearization.structure().mock_csc_transpose()
        )

        # the `damping` has the purpose of (optionally) improving conditioning
        self._damping: float = damping

    def solve(self, damping: Optional[float] = None, **kwargs) -> torch.Tensor:
        damping = damping or self._damping
        if not isinstance(self.linearization, SparseLinearization):
            raise RuntimeError(
                "CholmodSparseSolver only works with theseus.optimizer.SparseLinearization."
            )

        return CholmodSolveFunction.apply(
            self.linearization.A_val,
            self.linearization.b,
            self.linearization.structure(),
            self._symbolic_cholesky_decomposition,
            damping,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple, Union, cast

import torch

from theseus.core import CostFunction, CostWeight, Variable
from theseus.geometry import LieGroup, Vector


class DoubleIntegrator(CostFunction):
    def __init__(
        self,
        pose1: LieGroup,
        vel1: Vector,
        pose2: LieGroup,
        vel2: Vector,
        dt: Union[float, torch.Tensor, Variable],
        cost_weight: CostWeight,
        name: Optional[str] = None,
    ):
        super().__init__(cost_weight, name=name)
        dof = pose1.dof()
        if not (vel1.dof() == pose2.dof() == vel2.dof() == dof):
            raise ValueError(
                "All variables for a DoubleIntegrator must have the same dimension."
            )
        if not isinstance(dt, Variable):
            if not isinstance(dt, torch.Tensor):
                dt = torch.tensor(dt)
            self.dt = Variable(dt)
        else:
            self.dt = dt
        if self.dt.tensor.squeeze().ndim > 1:
            raise ValueError(
                "dt data must be a 0-D or 1-D tensor with numel in {1, batch_size}."
            )
        self.dt.tensor = self.dt.tensor.view(-1, 1)
        self.pose1 = pose1
        self.vel1 = vel1
        self.pose2 = pose2
        self.vel2 = vel2
        self.register_optim_vars(["pose1", "vel1", "pose2", "vel2"])
        self.register_aux_vars(["dt"])
        self.weight = cost_weight

    def dim(self):
        return 2 * self.pose1.dof()

    def _new_pose_diff(
        self, jacobians: Optional[List[torch.Tensor]] = None
    ) -> torch.Tensor:
        return self.pose1.local(self.pose2, jacobians=jacobians)

    def _error_from_pose_diff(self, pose_diff: torch.Tensor) -> torch.Tensor:
        pose_diff_err = pose_diff - self.dt.tensor.view(-1, 1) * self.vel1.tensor
        vel_diff = self.vel2.tensor - self.vel1.tensor
        return torch.cat([pose_diff_err, vel_diff], dim=1)

    def error(self) -> torch.Tensor:
        return self._error_from_pose_diff(self._new_pose_diff())

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        # Pre-allocate jacobian tensors
        batch_size = self.pose1.shape[0]
        dof = self.pose1.dof()
        dtype = self.pose1.dtype
        device = self.pose1.device
        Jerr_pose1 = torch.zeros(batch_size, 2 * dof, dof, dtype=dtype, device=device)
        Jerr_vel1 = torch.zeros_like(Jerr_pose1)
        Jerr_pose2 = torch.zeros_like(Jerr_pose1)
        Jerr_vel2 = torch.zeros_like(Jerr_pose1)

        Jlocal: List[torch.Tensor] = []
        error = self._error_from_pose_diff(self._new_pose_diff(Jlocal))
        Jerr_pose1[:, :dof, :] = Jlocal[0]
        identity = torch.eye(dof, dtype=dtype, device=device).repeat(batch_size, 1, 1)
        Jerr_vel1[:, :dof, :] = -self.dt.tensor.view(-1, 1, 1) * identity
        Jerr_vel1[:, dof:, :] = -identity
        Jerr_pose2[:, :dof, :] = Jlocal[1]
        Jerr_vel2[:, dof:, :] = identity
        return [Jerr_pose1, Jerr_vel1, Jerr_pose2, Jerr_vel2], error

    def _copy_impl(self, new_name: Optional[str] = None) -> "DoubleIntegrator":
        return DoubleIntegrator(
            self.pose1.copy(),
            self.vel1.copy(),
            self.pose2.copy(),
            self.vel2.copy(),
            self.dt.copy(),
            self.weight.copy(),
            name=new_name,
        )


class GPCostWeight(CostWeight):
    # Qc_inv is either a single square matrix or a batch of square matrices
    # dt is either a single element or a 1-D batch
    def __init__(
        self,
        Qc_inv: Union[Variable, torch.Tensor],
        dt: Union[float, Variable, torch.Tensor],
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        if not isinstance(dt, Variable):
            if not isinstance(dt, torch.Tensor):
                dt = torch.tensor(dt)
            dt = Variable(dt)
        if dt.tensor.squeeze().ndim > 1:
            raise ValueError("dt must be a 0-D or 1-D tensor.")
        self.dt = dt
        self.dt.tensor = self.dt.tensor.view(-1, 1)

        if not isinstance(Qc_inv, Variable):
            Qc_inv = Variable(Qc_inv)
        if Qc_inv.ndim not in [2, 3]:
            raise ValueError("Qc_inv must be a single matrix or a batch of matrices.")
        if not Qc_inv.shape[-2] == Qc_inv.shape[-1]:
            raise ValueError("Qc_inv must contain square matrices.")
        self.Qc_inv = Qc_inv
        self.Qc_inv.tensor = (
            Qc_inv.tensor if Qc_inv.ndim == 3 else Qc_inv.tensor.unsqueeze(0)
        )
        self.register_aux_vars(["Qc_inv", "dt"])

    def _compute_cost_weight(self) -> torch.Tensor:
        batch_size, dof, _ = self.Qc_inv.shape
        cost_weight = torch.empty(
            batch_size,
            2 * dof,
            2 * dof,
            dtype=self.Qc_inv.dtype,
            device=self.Qc_inv.device,
        )
        dt_data = self.dt.tensor.view(-1, 1, 1)
        Q11 = 12.0 * dt_data.pow(-3.0) * self.Qc_inv.tensor
        Q12 = -6.0 * dt_data.pow(-2.0) * self.Qc_inv.tensor
        Q22 = 4.0 * dt_data.reciprocal() * self.Qc_inv.tensor
        cost_weight[:, :dof, :dof] = Q11
        cost_weight[:, :dof:, dof:] = Q12
        cost_weight[:, dof:, :dof] = Q12
        cost_weight[:, dof:, dof:] = Q22
        return (
            torch.linalg.cholesky(cost_weight.transpose(-2, -1).conj())
            .transpose(-2, -1)
            .conj()
        )

    def weight_error(self, error: torch.Tensor) -> torch.Tensor:
        weights = self._compute_cost_weight()
        return torch.matmul(weights, error.unsqueeze(2)).squeeze(2)

    def weight_jacobians_and_error(
        self,
        jacobians: List[torch.Tensor],
        error: torch.Tensor,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        cost_weight = self._compute_cost_weight()
        error = torch.matmul(cost_weight, error.unsqueeze(2)).squeeze(2)
        new_jacobians = []
        for jac in jacobians:
            # Jacobian is batch_size x cost_function_dim x var_dim
            # This left multiplies the weights (inv cov.) to jacobian
            new_jacobians.append(torch.matmul(cost_weight, jac))
        return new_jacobians, error

    def _copy_impl(self, new_name: Optional[str] = None) -> "GPCostWeight":
        # need to pass data, since it could be Parameter(self.Qc_inv) in which
        # case Qc_inv won't be up to date.
        # This will change with the "learning happens outside API"
        return GPCostWeight(self.Qc_inv.copy(), self.dt.copy(), name=new_name)


class GPMotionModel(DoubleIntegrator):
    def __init__(
        self,
        pose1: LieGroup,
        vel1: Vector,
        pose2: LieGroup,
        vel2: Vector,
        dt: Union[float, Variable, torch.Tensor],
        cost_weight: GPCostWeight,
        name: Optional[str] = None,
    ):
        if not isinstance(cost_weight, GPCostWeight):
            raise ValueError(
                "GPMotionModel only accepts cost weights of type GPCostWeight. "
                "For other weight types, consider using DoubleIntegrator instead."
            )
        if not isinstance(dt, Variable):
            if not isinstance(dt, torch.Tensor):
                dt = torch.tensor(dt)
            self.dt = Variable(dt)
        else:
            self.dt = dt
        if self.dt.tensor.squeeze().ndim > 1:
            raise ValueError("dt must be a 0-D or 1-D tensor.")
        self.dt.tensor = self.dt.tensor.view(-1, 1)
        super().__init__(pose1, vel1, pose2, vel2, dt, cost_weight, name=name)

    def _copy_impl(self, new_name: Optional[str] = None) -> "GPMotionModel":
        return cast(GPMotionModel, super()._copy_impl(new_name=new_name))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import theseus as th

x = th.Point2()
y = th.Point2()

z1: th.Point2 = x + y  # noqa: F841
z2: th.Point2 = x - y  # noqa: F841
z3: th.Point2 = x * y  # noqa: F841
z4: th.Point2 = x / y  # noqa: F841
z5: th.Point2 = -x  # noqa: F841
z6: th.Point2 = x.cat(y)  # noqa: F841
z7: th.Point2 = x.abs()  # noqa: F841


x1 = th.Point3()
y1 = th.Point3()

w1: th.Point3 = x1 + y1  # noqa: F841
w2: th.Point3 = x1 - y1  # noqa: F841
w3: th.Point3 = x1 * y1  # noqa: F841
w4: th.Point3 = x1 / y1  # noqa: F841
w5: th.Point3 = -x1  # noqa: F841
w6: th.Point3 = x1.cat(y1)  # noqa: F841
w7: th.Point3 = x1.abs()  # noqa: F841



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple

import torch

import theseus as th


class Reprojection(th.CostFunction):
    def __init__(
        self,
        camera_pose: th.SE3,
        world_point: th.Point3,
        image_feature_point: th.Point2,
        focal_length: th.Vector,
        calib_k1: th.Vector = None,
        calib_k2: th.Vector = None,
        weight: Optional[th.CostWeight] = None,
        name: Optional[str] = None,
    ):
        if weight is None:
            weight = th.ScaleCostWeight(torch.tensor(1.0).to(dtype=camera_pose.dtype))
        super().__init__(
            cost_weight=weight,
            name=name,
        )
        self.camera_pose = camera_pose
        self.focal_length = focal_length
        self.calib_k1 = calib_k1
        self.calib_k2 = calib_k2
        batch_size = self.camera_pose.shape[0]
        if self.calib_k1 is None:
            self.calib_k1 = th.Vector(
                tensor=torch.zeros((batch_size, 1), dtype=camera_pose.dtype),
                name="calib_k1",
            )
        if self.calib_k2 is None:
            self.calib_k2 = th.Vector(
                tensor=torch.zeros((batch_size, 1), dtype=camera_pose.dtype),
                name="calib_k2",
            )
        self.world_point = world_point
        self.image_feature_point = image_feature_point

        self.register_optim_vars(["camera_pose", "world_point"])
        self.register_aux_vars(
            ["focal_length", "image_feature_point", "calib_k1", "calib_k2"]
        )

    def error(self) -> torch.Tensor:
        point_cam = self.camera_pose.transform_from(self.world_point)
        proj = -point_cam[:, :2] / point_cam[:, 2:3]
        proj_sqn = (proj * proj).sum(dim=1).unsqueeze(1)
        proj_factor = self.focal_length.tensor * (
            1.0 + proj_sqn * (self.calib_k1.tensor + proj_sqn * self.calib_k2.tensor)
        )
        point_projection = proj * proj_factor

        err = point_projection - self.image_feature_point.tensor
        return err

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        cpose_wpt_jacs: List[torch.Tensor] = []
        point_cam = self.camera_pose.transform_from(self.world_point, cpose_wpt_jacs)
        J = torch.cat(cpose_wpt_jacs, dim=2)

        proj = -point_cam[:, :2] / point_cam[:, 2:3]
        proj_sqn = (proj * proj).sum(dim=1).unsqueeze(1)
        proj_factor = self.focal_length.tensor * (
            1.0 + proj_sqn * (self.calib_k1.tensor + proj_sqn * self.calib_k2.tensor)
        )
        d_proj_factor = self.focal_length.tensor * (
            self.calib_k1.tensor + 2.0 * proj_sqn * self.calib_k2.tensor
        )
        point_projection = proj * proj_factor

        # derivative of N/D is (N' - ND'/D) / D
        d_num = J[:, 0:2, :]
        num_dden_den = torch.bmm(
            point_cam[:, :2].unsqueeze(2),
            (J[:, 2, :] / point_cam[:, 2:3]).unsqueeze(1),
        )
        proj_jac = (num_dden_den - d_num) / point_cam[:, 2:].unsqueeze(2)
        proj_sqn_jac = 2.0 * proj.unsqueeze(2) * torch.bmm(proj.unsqueeze(1), proj_jac)
        point_projection_jac = proj_jac * proj_factor.unsqueeze(
            2
        ) + proj_sqn_jac * d_proj_factor.unsqueeze(2)

        err = point_projection - self.image_feature_point.tensor
        return [point_projection_jac[..., :6], point_projection_jac[..., 6:]], err

    def dim(self) -> int:
        return 2

    def to(self, *args, **kwargs):
        super().to(*args, **kwargs)

    def _copy_impl(self, new_name: Optional[str] = None) -> "Reprojection":
        return Reprojection(
            self.camera_pose.copy(),
            self.world_point.copy(),
            self.image_feature_point.copy(),
            self.focal_length.copy(),
            calib_k1=self.calib_k1.copy(),
            calib_k2=self.calib_k2.copy(),
            weight=self.weight.copy(),
            name=new_name,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy

import pytest  # noqa: F401
import torch

import theseus as th

from .common import (
    check_another_theseus_function_is_copy,
    check_another_theseus_tensor_is_copy,
)


def test_copy_scale_cost_weight():
    scale = th.Variable(torch.tensor(1.0))
    p1 = th.ScaleCostWeight(scale, name="scale_cost_weight")
    for the_copy in [p1.copy(), copy.deepcopy(p1)]:
        check_another_theseus_function_is_copy(p1, the_copy, new_name=f"{p1.name}_copy")
        check_another_theseus_tensor_is_copy(scale, the_copy.scale)


def test_copy_diagonal_cost_weight():
    diagonal = th.Variable(torch.ones(3))
    p1 = th.DiagonalCostWeight(diagonal, name="diagonal_cost_weight")
    for the_copy in [p1.copy(), copy.deepcopy(p1)]:
        check_another_theseus_function_is_copy(p1, the_copy, new_name=f"{p1.name}_copy")
        check_another_theseus_tensor_is_copy(diagonal, the_copy.diagonal)


def test_scale_cost_weight():
    for dim in [1, 2, 10]:
        for batch_size in [1, 2, 10]:
            v1 = th.Vector(tensor=torch.ones(batch_size, dim))
            z = th.Vector(tensor=torch.zeros(batch_size, dim))
            scale = torch.randn(1).item()
            expected_err = torch.ones(batch_size, dim) * scale
            expected_jac = (torch.eye(dim).unsqueeze(0) * scale).expand(
                batch_size, dim, dim
            )

            def _check(cw):
                cf1 = th.Difference(v1, z, cw)
                jacobians, error = cf1.weighted_jacobians_error()
                assert error.allclose(expected_err)
                assert jacobians[0].allclose(expected_jac)

            _check(th.ScaleCostWeight(scale))
            _check(th.ScaleCostWeight(torch.ones(1) * scale))
            _check(th.ScaleCostWeight(torch.ones(batch_size) * scale))
            _check(th.ScaleCostWeight(torch.ones(1, 1) * scale))
            _check(th.ScaleCostWeight(torch.ones(batch_size, 1) * scale))
            _check(th.ScaleCostWeight(th.Variable(torch.ones(1, 1) * scale)))

            batched_scale = torch.randn(batch_size, 1)
            expected_err = batched_scale * torch.ones(batch_size, dim)
            expected_jac = (torch.eye(dim).unsqueeze(0)).expand(
                batch_size, dim, dim
            ) * batched_scale.view(-1, 1, 1)

            cf1 = th.Difference(v1, z, th.ScaleCostWeight(batched_scale))
            jacobians, error = cf1.weighted_jacobians_error()
            assert error.allclose(expected_err)
            assert jacobians[0].allclose(expected_jac)


def test_diagonal_cost_weight():
    for dim in [1, 2, 10]:
        for batch_size in [1, 2, 10]:
            v1 = th.Vector(tensor=torch.ones(batch_size, dim))
            z = th.Vector(tensor=torch.zeros(batch_size, dim))
            diagonal = torch.randn(dim)
            d_matrix = diagonal.diag().unsqueeze(0)
            expected_err = (d_matrix @ torch.ones(batch_size, dim, 1)).view(
                batch_size, dim
            )
            expected_jac = d_matrix @ (torch.eye(dim).unsqueeze(0)).expand(
                batch_size, dim, dim
            )

            def _check(cw):
                cf1 = th.Difference(v1, z, cw)
                jacobians, error = cf1.weighted_jacobians_error()
                assert error.allclose(expected_err)
                assert jacobians[0].allclose(expected_jac)

            _check(th.DiagonalCostWeight(diagonal.tolist()))
            _check(th.DiagonalCostWeight(diagonal))
            _check(th.DiagonalCostWeight(diagonal.view(1, dim)))
            _check(th.DiagonalCostWeight(th.Variable(diagonal)))

            batched_diagonal = torch.randn(batch_size, dim)
            d_matrix = batched_diagonal.diag_embed()
            expected_err = (d_matrix @ torch.ones(batch_size, dim, 1)).view(
                batch_size, dim
            )
            expected_jac = d_matrix @ (torch.eye(dim).unsqueeze(0)).expand(
                batch_size, dim, dim
            )

            cf1 = th.Difference(v1, z, th.DiagonalCostWeight(batched_diagonal))
            jacobians, error = cf1.weighted_jacobians_error()
            assert error.allclose(expected_err)
            assert jacobians[0].allclose(expected_jac)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest  # noqa: F401

import theseus as th


@pytest.mark.parametrize(
    "var_type", [(th.Vector, 1), (th.Vector, 2), (th.SE2, None), (th.SE3, None)]
)
@pytest.mark.parametrize("batch_size", [1, 10])
def test_state_history(var_type, batch_size):
    cls_, dof = var_type

    rand_args = (batch_size, dof) if cls_ == th.Vector else (batch_size,)
    v1 = cls_(tensor=cls_.rand(*rand_args).tensor, name="v1")
    v2 = cls_(tensor=cls_.rand(*rand_args).tensor, name="v2")
    w = th.ScaleCostWeight(1.0)

    objective = th.Objective()
    objective.add(th.Difference(v1, v2, w))

    max_iters = 10
    optimizer = th.GaussNewton(objective, max_iterations=max_iters)

    info = optimizer.optimize(track_state_history=True)

    for var in objective.optim_vars.values():
        assert var.name in info.state_history
        assert info.state_history[var.name].shape == (objective.batch_size,) + tuple(
            var.shape[1:]
        ) + (max_iters + 1,)



#!/usr/bin/env python
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
import setuptools
import sys

try:
    import torch
    from torch.utils import cpp_extension as torch_cpp_ext
except ModuleNotFoundError:
    print("Theseus installation requires torch.")
    sys.exit(1)


def parse_requirements_file(path):
    with open(path) as f:
        reqs = []
        for line in f:
            line = line.strip()
            reqs.append(line.split("==")[0])
    return reqs


reqs_main = parse_requirements_file("requirements/main.txt")
reqs_dev = parse_requirements_file("requirements/dev.txt")
root_dir = Path(__file__).parent

with open(Path("theseus") / "__init__.py", "r") as f:
    for line in f:
        if "__version__" in line:
            version = line.split("__version__ = ")[1].rstrip().strip('"')

with open("README.md", "r") as fh:
    long_description = fh.read()

if torch.cuda.is_available():
    ext_modules = [
        torch_cpp_ext.CUDAExtension(
            name="theseus.extlib.mat_mult",
            sources=[str(root_dir / "theseus" / "extlib" / "mat_mult.cu")],
        ),
        torch_cpp_ext.CUDAExtension(
            name="theseus.extlib.cusolver_lu_solver",
            sources=[
                str(root_dir / "theseus" / "extlib" / "cusolver_lu_solver.cpp"),
                str(root_dir / "theseus" / "extlib" / "cusolver_sp_defs.cpp"),
            ],
            include_dirs=[str(root_dir)],
            libraries=["cusolver"],
        ),
    ]
else:
    print("No CUDA support found. CUDA extensions won't be installed.")
    ext_modules = []

setuptools.setup(
    name="theseus-ai",
    version=version,
    author="Meta Research",
    description="A library for differentiable nonlinear optimization.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/facebookresearch/theseus",
    keywords="differentiable optimization, nonlinear least squares, factor graphs",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.7",
    install_requires=reqs_main,
    extras_require={"dev": reqs_main + reqs_dev},
    cmdclass={"build_ext": torch_cpp_ext.BuildExtension},
    ext_modules=ext_modules,
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import Callable, List, Optional, Sequence, Tuple, cast

import torch
import torch.autograd.functional as autogradF
from typing_extensions import Protocol

from theseus.geometry import Manifold

from .cost_weight import CostWeight, ScaleCostWeight
from .theseus_function import TheseusFunction
from .variable import Variable


# A cost function is defined by the variables interacting in it,
# and a cost weight for weighting errors and jacobians
# This is an abstract class for cost functions of different types.
# Each concrete function must implement an error method and the
# jacobian of the error, by implementing abstract methods
# `error` and `jacobians`, respectively.
class CostFunction(TheseusFunction, abc.ABC):
    def __init__(
        self,
        cost_weight: CostWeight,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        self._weight = cost_weight

    @property
    def weight(self) -> CostWeight:
        return self._weight

    @weight.setter
    def weight(self, weight: CostWeight):
        self._weight = weight

    @abc.abstractmethod
    def error(self) -> torch.Tensor:
        pass

    # Returns (jacobians, error)
    @abc.abstractmethod
    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        pass

    @abc.abstractmethod
    def dim(self) -> int:
        pass

    def weighted_error(self) -> torch.Tensor:
        error = self.error()
        return self.weight.weight_error(error)

    def weighted_jacobians_error(
        self,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        jacobian, err = self.jacobians()
        return self.weight.weight_jacobians_and_error(jacobian, err)

    # Must copy everything
    @abc.abstractmethod
    def _copy_impl(self, new_name: Optional[str] = None) -> "CostFunction":
        pass

    # Here for mypy compatibility downstream
    def copy(
        self, new_name: Optional[str] = None, keep_variable_names: bool = False
    ) -> "CostFunction":
        return cast(
            CostFunction,
            super().copy(new_name=new_name, keep_variable_names=keep_variable_names),
        )

    # calls to() on the cost weight, variables and any internal tensors
    def to(self, *args, **kwargs):
        super().to(*args, **kwargs)
        self.weight.to(*args, **kwargs)


# Function protocol for learnable cost functions. `optim_vars` and `aux_vars` are
# separated to facilitate internal autograd handling of optimization variables
class ErrFnType(Protocol):
    def __call__(
        self, optim_vars: Tuple[Variable, ...], aux_vars: Tuple[Variable, ...]
    ) -> torch.Tensor:
        ...


# The error function is assumed to receive variables in the format
#    err_fn(
#       optim_vars=(optim_vars[0].tensor, ..., optim_vars[N - 1].tensor),
#       aux_vars=(aux_vars[0].tensor, ..., aux_vars[M -1].tensor)
#   )
#
# The user also needs to explicitly specify the output's dimension
class AutoDiffCostFunction(CostFunction):
    def __init__(
        self,
        optim_vars: Sequence[Manifold],
        err_fn: ErrFnType,
        dim: int,
        cost_weight: Optional[CostWeight] = None,
        aux_vars: Optional[Sequence[Variable]] = None,
        name: Optional[str] = None,
        autograd_strict: bool = False,
        autograd_vectorize: bool = False,
        autograd_loop_over_batch: bool = False,
    ):
        if cost_weight is None:
            cost_weight = ScaleCostWeight(1.0)
        super().__init__(cost_weight, name=name)
        # this avoids doing aux_vars=[], which is a bad default since [] is mutable
        aux_vars = aux_vars or []

        if len(optim_vars) < 1:
            raise ValueError(
                "AutodiffCostFunction must receive at least one optimization variable."
            )
        self.register_vars(optim_vars, is_optim_vars=True)
        self.register_vars(aux_vars, is_optim_vars=False)

        self._err_fn = err_fn
        self._dim = dim
        self._autograd_strict = autograd_strict
        self._autograd_vectorize = autograd_vectorize

        # The following are auxiliary Variable objects to hold tensor data
        # during jacobian computation without modifying the original Variable objects
        self._tmp_optim_vars = tuple(v.copy() for v in optim_vars)

        self._tmp_optim_vars_for_loop = None
        self._tmp_aux_vars_for_loop = None

        if autograd_loop_over_batch:
            self._tmp_optim_vars_for_loop = tuple(v.copy() for v in optim_vars)
            self._tmp_aux_vars_for_loop = tuple(v.copy() for v in aux_vars)

            for i, optim_var in enumerate(optim_vars):
                self._tmp_optim_vars_for_loop[i].update(optim_var.tensor)

            for i, aux_var in enumerate(aux_vars):
                self._tmp_aux_vars_for_loop[i].update(aux_var.tensor)

        self._autograd_loop_over_batch = autograd_loop_over_batch

    def _compute_error(
        self,
    ) -> Tuple[torch.Tensor, Tuple[Manifold, ...], Tuple[Variable, ...]]:
        optim_vars = tuple(v for v in self.optim_vars)
        aux_vars = tuple(v for v in self.aux_vars)
        err = self._err_fn(optim_vars=optim_vars, aux_vars=aux_vars)
        if err.shape[1] != self.dim():
            raise ValueError(
                "Output dimension of given error function doesn't match self.dim()."
            )
        return err, optim_vars, aux_vars

    def error(self) -> torch.Tensor:
        return self._compute_error()[0]

    def _make_jac_fn(
        self, tmp_optim_vars: Tuple[Manifold, ...], tmp_aux_vars: Tuple[Variable, ...]
    ) -> Callable:
        def jac_fn(*optim_vars_tensors_):
            assert len(optim_vars_tensors_) == len(tmp_optim_vars)
            for i, tensor in enumerate(optim_vars_tensors_):
                tmp_optim_vars[i].update(tensor)

            return self._err_fn(optim_vars=tmp_optim_vars, aux_vars=tmp_aux_vars)

        return jac_fn

    def _compute_autograd_jacobian(
        self, optim_tensors: Tuple[torch.Tensor, ...], jac_fn: Callable
    ) -> Tuple[torch.Tensor, ...]:
        return autogradF.jacobian(
            jac_fn,
            optim_tensors,
            create_graph=True,
            strict=self._autograd_strict,
            vectorize=self._autograd_vectorize,
        )

    # Returns (jacobians, error)
    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        err, optim_vars, aux_vars = self._compute_error()

        if self._autograd_loop_over_batch:
            jacobians_raw_loop: List[Tuple[torch.Tensor, ...]] = []
            for n in range(optim_vars[0].shape[0]):
                for i, aux_var in enumerate(aux_vars):
                    self._tmp_aux_vars_for_loop[i].update(aux_var.tensor[n : n + 1])

                jacobians_n = self._compute_autograd_jacobian(
                    tuple(v.tensor[n : n + 1] for v in optim_vars),
                    self._make_jac_fn(
                        self._tmp_optim_vars_for_loop, self._tmp_aux_vars_for_loop
                    ),
                )
                jacobians_raw_loop.append(jacobians_n)

            jacobians_full = [
                torch.cat([jac[k][:, :, 0, :] for jac in jacobians_raw_loop], dim=0)
                for k in range(len(optim_vars))
            ]
        else:
            jacobians_raw = self._compute_autograd_jacobian(
                tuple(v.tensor for v in optim_vars),
                self._make_jac_fn(self._tmp_optim_vars, aux_vars),
            )
            aux_idx = torch.arange(err.shape[0])  # batch_size
            jacobians_full = [jac[aux_idx, :, aux_idx, :] for jac in jacobians_raw]

        # torch autograd returns shape (batch_size, dim, batch_size, var_dim), which
        # includes derivatives of batches against each other.
        # this indexing recovers only the derivatives wrt the same batch
        jacobians = list(
            v.project(jac, is_sparse=True) for v, jac in zip(optim_vars, jacobians_full)
        )

        return jacobians, err

    def dim(self) -> int:
        return self._dim

    def _copy_impl(self, new_name: Optional[str] = None) -> "AutoDiffCostFunction":
        return AutoDiffCostFunction(
            [v.copy() for v in self.optim_vars],
            self._err_fn,
            self._dim,
            aux_vars=[v.copy() for v in self.aux_vars],
            cost_weight=self.weight.copy(),
            name=new_name,
        )

    def to(self, *args, **kwargs):
        # calls to() on the cost weight, variables and any internal tensors
        super().to(*args, **kwargs)
        for var in self._tmp_optim_vars:
            var.to(*args, **kwargs)

        if self._autograd_loop_over_batch:
            for var in self._tmp_optim_vars_for_loop:
                var.to(*args, **kwargs)

            for var in self._tmp_aux_vars_for_loop:
                var.to(*args, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

from .dense_solver import CholeskyDenseSolver, DenseSolver, LUDenseSolver
from .linear_optimizer import LinearOptimizer
from .linear_solver import LinearSolver
from .lu_cuda_sparse_solver import LUCudaSparseSolver
from .sparse_solver import CholmodSparseSolver



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th

from .common import (
    MockCostFunction,
    MockCostWeight,
    MockVar,
    NullCostWeight,
    check_another_theseus_tensor_is_copy,
    create_mock_cost_functions,
    create_objective_with_mock_cost_functions,
)


def test_add():
    objective, cost_functions, names, *_ = create_objective_with_mock_cost_functions()

    for cost_function, name in zip(cost_functions, names):
        assert name in objective.cost_functions
        assert objective.cost_functions[name] == cost_function

    # --------- All the code that follows test for handling of errors ---------
    def _create_cost_function_with_n_vars_and_m_aux(
        cost_function_name, var_names, aux_names, cost_weight
    ):
        the_vars = [MockVar(1, name=name_) for name_ in var_names]
        the_aux = [MockVar(1, name=name_) for name_ in aux_names]
        return MockCostFunction(
            the_vars,
            the_aux,
            cost_weight,
            name=cost_function_name,
        )

    # ###  Testing no duplicate names allowed ###
    # duplicate cost function name
    new_cost_function = _create_cost_function_with_n_vars_and_m_aux(
        cost_functions[0].name, ["none1", "none2"], [], NullCostWeight()
    )
    with pytest.raises(ValueError):
        objective.add(new_cost_function)
    # duplicate variable name
    new_cost_function = _create_cost_function_with_n_vars_and_m_aux(
        "new", [cost_functions[0].optim_var_0.name], [], NullCostWeight()
    )
    with pytest.raises(ValueError):
        objective.add(new_cost_function)

    # This for loop tests two failure cases:
    # i) duplicate aux. variable name and
    # ii) same name used both for optimization var and aux. variable
    for aux_name in [
        cost_functions[0].aux_var_0.name,
        cost_functions[0].optim_var_0.name,
    ]:
        new_cost_function = _create_cost_function_with_n_vars_and_m_aux(
            f"another_{aux_name}", ["some_name"], [aux_name], NullCostWeight()
        )
        with pytest.raises(ValueError):
            objective.add(new_cost_function)

    # -------------- Test similar stuff but for cost weight --------
    cost_weight = MockCostWeight(
        th.Variable(torch.ones(1), name="cost_weight_aux"),
        add_dummy_var_with_name="cost_weight_var_1",
    )
    yet_another_cost_function = _create_cost_function_with_n_vars_and_m_aux(
        "yet_another", ["yet_var_1"], ["yet_aux_1"], cost_weight
    )
    with pytest.raises(RuntimeError):  # optim var associated to weight
        objective.add(yet_another_cost_function)  # no conflict here

    cost_weight_with_conflict_in_aux_var = MockCostWeight(
        th.Variable(torch.ones(1), name="cost_weight_aux")
    )
    cost_function_with_cost_weight_conflict = (
        _create_cost_function_with_n_vars_and_m_aux(
            "yet_another_2",
            ["yet_var_2"],
            ["yet_aux_2"],
            cost_weight_with_conflict_in_aux_var,
        )
    )
    with pytest.raises(ValueError):
        objective.add(cost_function_with_cost_weight_conflict)

    cost_weight_with_conflict_in_var = MockCostWeight(
        th.Variable(torch.ones(1), name="cost_weight_aux"),
        add_dummy_var_with_name="cost_weight_var_1",
    )
    another_cost_function_with_cost_weight_conflict = (
        _create_cost_function_with_n_vars_and_m_aux(
            "yet_another_3",
            ["yet_var_3"],
            ["yet_aux_3"],
            cost_weight_with_conflict_in_var,
        )
    )
    with pytest.raises(ValueError):
        objective.add(another_cost_function_with_cost_weight_conflict)


def test_getters():
    (
        objective,
        cost_functions,
        names,
        var_to_cost_functions,
        aux_to_cost_functions,
    ) = create_objective_with_mock_cost_functions()
    for cost_function, name in zip(cost_functions, names):
        assert id(objective.get_cost_function(name)) == id(cost_function)

    for var in var_to_cost_functions:
        assert id(var) == id(objective.get_optim_var(var.name))

    for aux in aux_to_cost_functions:
        assert id(aux) == id(objective.get_aux_var(aux.name))

    assert objective.get_cost_function("bad-name") is None
    assert objective.get_optim_var("bad-name") is None
    assert objective.get_aux_var("bad-name") is None


def test_has_cost_function_and_has_var():
    objective, cost_functions, names, *_ = create_objective_with_mock_cost_functions()

    for name in names:
        assert objective.has_cost_function(name)

    for cost_function in cost_functions:
        for var in cost_function.optim_vars:
            assert objective.has_optim_var(var.name)

    for cost_function in cost_functions:
        for aux in cost_function.aux_vars:
            assert objective.has_aux_var(aux.name)


def test_add_and_erase_step_by_step():
    var1 = MockVar(1, tensor=None, name="var1")
    var2 = MockVar(1, tensor=None, name="var2")
    var3 = MockVar(1, tensor=None, name="var3")
    aux1 = MockVar(1, tensor=None, name="aux1")
    aux2 = MockVar(1, tensor=None, name="aux2")
    cw1 = MockCostWeight(
        aux1, name="cw1"
    )  # , add_dummy_var_with_name="ignored_optim_var")
    cw2 = MockCostWeight(aux2, name="cw2")  # , add_optim_var=var1)

    cf1 = MockCostFunction([var1, var2], [aux1, aux2], cw1, name="cf1")
    cf2 = MockCostFunction([var1, var3], [aux1], cw2, name="cf2")
    cf3 = MockCostFunction([var2, var3], [aux2], cw2, name="cf3")

    objective = th.Objective()
    for cost_function in [cf1, cf2, cf3]:
        objective.add(cost_function)

    for name in ["var1", "var2", "var2"]:
        assert name in objective.optim_vars
    for name in ["aux1", "aux2"]:
        assert name in objective.aux_vars

    # Checks that the objective is maintaining variable -> fn correctly, against
    # a list of expected functions
    def _check_funs_for_variable(
        var_, expected_fn_list_, optim_var=True, is_cost_weight_optim=False
    ):
        # If var should have been deleted, check that no trace of it remains
        if not expected_fn_list_:
            if optim_var:
                assert var_ not in objective.functions_for_optim_vars
                if is_cost_weight_optim:
                    assert var_ not in objective.cost_weight_optim_vars
                else:
                    assert var_ not in objective.optim_vars
            else:
                assert var_ not in objective.functions_for_aux_vars
                assert var_ not in objective.aux_vars
            return

        # Otherwise check that list of connected functions match the expected
        if optim_var:
            obj_container = objective.functions_for_optim_vars[var_]
        else:
            obj_container = objective.functions_for_aux_vars[var_]
        assert len(obj_container) == len(expected_fn_list_)
        for fun_ in expected_fn_list_:
            assert fun_ in obj_container

    # Runs the above check for all variables in this graph, given expected lists
    # for each
    def _check_all_vars(v1_lis_, v2_lis_, v3_lis_, cw1_opt_lis_, a1_lis_, a2_lis_):
        _check_funs_for_variable(var1, v1_lis_)
        _check_funs_for_variable(var2, v2_lis_)
        _check_funs_for_variable(var3, v3_lis_)
        _check_funs_for_variable(aux1, a1_lis_, optim_var=False)
        _check_funs_for_variable(aux2, a2_lis_, optim_var=False)

    v1_lis = [cf1, cf2]
    v2_lis = [cf1, cf3]
    v3_lis = [cf2, cf3]
    cw1o_lis = [cw1]
    a1_lis = [cw1, cf1, cf2]
    a2_lis = [cw2, cf1, cf3]

    _check_all_vars(v1_lis, v2_lis, v3_lis, cw1o_lis, a1_lis, a2_lis)

    objective.erase("cf1")
    v1_lis = [cf2]
    v2_lis = [cf3]
    cw1o_lis = []
    a1_lis = [cf2]  # cf1 and cw1 are deleted, since cw1 not used by any other cost fn
    a2_lis = [cw2, cf3]
    _check_all_vars(v1_lis, v2_lis, v3_lis, cw1o_lis, a1_lis, a2_lis)
    assert cw1 not in objective.cost_functions_for_weights

    objective.erase("cf2")
    v1_lis = []
    v3_lis = [cf3]
    a1_lis = []
    _check_all_vars(v1_lis, v2_lis, v3_lis, cw1o_lis, a1_lis, a2_lis)
    assert cw2 in objective.cost_functions_for_weights

    objective.erase("cf3")
    v1_lis = []
    v2_lis = []
    v3_lis = []
    a2_lis = []
    _check_all_vars(v1_lis, v2_lis, v3_lis, cw1o_lis, a1_lis, a2_lis)
    assert cw2 not in objective.cost_functions_for_weights


def test_objective_error():
    def _check_error_for_data(v1_data_, v2_data_, error_, error_type):
        expected_error = torch.cat([v1_data_, v2_data_], dim=1) * w

        if error_type == "error":
            assert error_.allclose(expected_error)
        else:
            assert error_.allclose(expected_error.norm(dim=1) ** 2)

    def _check_variables(objective, input_tensors, v1_data, v2_data, also_update):

        if also_update:
            assert objective.optim_vars["v1"].tensor is input_tensors["v1"]
            assert objective.optim_vars["v2"].tensor is input_tensors["v2"]
        else:
            assert objective.optim_vars["v1"].tensor is not input_tensors["v1"]
            assert objective.optim_vars["v2"].tensor is not input_tensors["v2"]

            assert objective.optim_vars["v1"].tensor is v1_data
            assert objective.optim_vars["v2"].tensor is v2_data

    def _check_error_and_variables(
        v1_data_, v2_data_, error_, error_type, objective, input_tensors, also_update
    ):

        _check_error_for_data(v1_data_, v2_data_, error_, error_type)

        _check_variables(objective, input_tensors, v1_data, v2_data, also_update)

    for _ in range(10):
        f1, f2 = np.random.random(), np.random.random()
        dof = np.random.randint(2, 10)
        batch_size = np.random.randint(2, 10)
        v1 = th.Vector(dof=dof, name="v1")
        v2 = th.Vector(dof=dof, name="v2")
        z = th.Vector(tensor=torch.zeros(batch_size, dof), name="z")

        w = np.random.random()
        # This cost functions will just compute the norm of each vector, scaled by w
        weight = th.ScaleCostWeight(w)
        d1 = th.Difference(v1, z, weight, name="d1")
        d2 = th.Difference(v2, z, weight, name="d2")

        objective = th.Objective()
        objective.add(d1)
        objective.add(d2)

        v1_data = torch.ones(batch_size, dof) * f1
        v2_data = torch.ones(batch_size, dof) * f2
        objective.update({"v1": v1_data, "v2": v2_data})
        error = objective.error()
        _check_error_for_data(v1_data, v2_data, error, "error")
        error_norm_2 = objective.error_squared_norm()

        assert error.shape == (batch_size, 2 * dof)
        _check_error_for_data(v1_data, v2_data, error_norm_2, "error_norm_2")

        v1_data_new = torch.ones(batch_size, dof) * f1 * 0.1
        v2_data_new = torch.ones(batch_size, dof) * f2 * 0.1

        input_tensors = {"v1": v1_data_new, "v2": v2_data_new}

        error = objective.error(input_tensors=input_tensors, also_update=False)

        _check_error_and_variables(
            v1_data_new,
            v2_data_new,
            error,
            "error",
            objective,
            input_tensors,
            also_update=False,
        )

        v1_data_new = torch.ones(batch_size, dof) * f1 * 0.3
        v2_data_new = torch.ones(batch_size, dof) * f2 * 0.3

        input_tensors = {"v1": v1_data_new, "v2": v2_data_new}

        error_norm_2 = objective.error_squared_norm(
            input_tensors=input_tensors, also_update=False
        )

        _check_error_and_variables(
            v1_data_new,
            v2_data_new,
            error_norm_2,
            "error_norm_2",
            objective,
            input_tensors,
            also_update=False,
        )

        v1_data_new = torch.ones(batch_size, dof) * f1 * 0.4
        v2_data_new = torch.ones(batch_size, dof) * f2 * 0.4

        input_tensors = {"v1": v1_data_new, "v2": v2_data_new}

        error = objective.error(input_tensors=input_tensors, also_update=True)

        _check_error_and_variables(
            v1_data_new,
            v2_data_new,
            error,
            "error",
            objective,
            input_tensors,
            also_update=True,
        )

        v1_data_new = torch.ones(batch_size, dof) * f1 * 0.4
        v2_data_new = torch.ones(batch_size, dof) * f2 * 0.4

        input_tensors = {"v1": v1_data_new, "v2": v2_data_new}

        error_norm_2 = objective.error_squared_norm(
            input_tensors=input_tensors, also_update=True
        )

        _check_error_and_variables(
            v1_data_new,
            v2_data_new,
            error_norm_2,
            "error_norm_2",
            objective,
            input_tensors,
            also_update=True,
        )


def test_get_cost_functions_connected_to_vars():
    (
        objective,
        _,
        _,
        optim_var_to_cost_functions,
        aux_to_cost_functions,
    ) = create_objective_with_mock_cost_functions()

    def _check_connections(var_to_cost_fns, objective_get_var_method):
        for variable, expected_cost_functions in var_to_cost_fns.items():

            def _check_cost_functions(cost_functions):
                assert len(cost_functions) == len(expected_cost_functions)
                for cost_function in cost_functions:
                    assert cost_function in expected_cost_functions

            _check_cost_functions(objective_get_var_method(variable))
            _check_cost_functions(objective_get_var_method(variable.name))

            with pytest.raises(ValueError):
                objective_get_var_method("var_not_in_objective")
                objective_get_var_method(MockVar(None, name="none"))

    _check_connections(
        optim_var_to_cost_functions, objective.get_functions_connected_to_optim_var
    )
    _check_connections(
        aux_to_cost_functions, objective.get_functions_connected_to_aux_var
    )


def test_copy():
    objective, cost_functions, *_ = create_objective_with_mock_cost_functions(
        torch.ones(1, 1), MockCostWeight(th.Variable(torch.ones(1)))
    )
    new_objective = objective.copy()
    assert new_objective is not objective

    assert objective.size_cost_functions() == new_objective.size_cost_functions()
    for cost_function, new_cost_function in zip(
        cost_functions, new_objective.cost_functions.values()
    ):
        assert cost_function.num_optim_vars() == new_cost_function.num_optim_vars()
        assert cost_function.num_aux_vars() == new_cost_function.num_aux_vars()
        assert cost_function.name == new_cost_function.name
        assert cost_function is not new_objective
        for var, new_var in zip(cost_function.optim_vars, new_cost_function.optim_vars):
            check_another_theseus_tensor_is_copy(var, new_var)
        for aux, new_aux in zip(cost_function.aux_vars, new_cost_function.aux_vars):
            check_another_theseus_tensor_is_copy(aux, new_aux)


def test_copy_no_duplicate_cost_weights():
    objective = th.Objective()
    v1 = MockVar(1, name="v1")
    cw1 = th.ScaleCostWeight(1.0)
    cw2 = th.ScaleCostWeight(2.0)
    objective.add(MockCostFunction([v1], [], cw1, "cf1"))
    objective.add(MockCostFunction([v1], [], cw1, "cf2"))
    objective.add(MockCostFunction([v1], [], cw2, "cf3"))
    objective.add(MockCostFunction([v1], [], cw2, "cf4"))
    objective.add(MockCostFunction([v1], [], cw2, "cf5"))

    objective_copy = objective.copy()
    seen_cw1 = set()
    seen_cw2 = set()
    for cf in objective_copy.cost_functions.values():
        if cf.name in ["cf1", "cf2"]:
            scale = 1.0
            original_weight = cw1
            set_to_add = seen_cw1
        else:
            scale = 2.0
            original_weight = cw2
            set_to_add = seen_cw2

        assert isinstance(cf.weight, th.ScaleCostWeight)
        assert cf.weight.scale.tensor.item() == scale
        assert cf.weight is not original_weight
        set_to_add.add(cf.weight)
    assert len(seen_cw1) == 1
    assert len(seen_cw2) == 1


def test_update_updates_properly():
    (
        objective,
        _,
        _,
        var_to_cost_functions,
        aux_to_cost_functions,
    ) = create_objective_with_mock_cost_functions(
        torch.ones(1, 1),
        MockCostWeight(th.Variable(torch.ones(1), name="cost_weight_aux")),
    )

    input_tensors = {}
    for var in var_to_cost_functions:
        input_tensors[var.name] = 2 * var.tensor.clone()
    for aux in aux_to_cost_functions:
        input_tensors[aux.name] = 2 * aux.tensor.clone()

    objective.update(input_tensors=input_tensors)
    assert objective.batch_size == 1

    for var_name, data in input_tensors.items():
        if var_name in [v.name for v in var_to_cost_functions]:
            var_ = objective.get_optim_var(var_name)
        if var_name in [aux.name for aux in aux_to_cost_functions]:
            var_ = objective.get_aux_var(var_name)
        assert data is var_.tensor


def test_update_raises_batch_size_error():
    (
        objective,
        _,
        _,
        var_to_cost_functions,
        aux_to_cost_functions,
    ) = create_objective_with_mock_cost_functions(
        torch.ones(1, 1),
        MockCostWeight(th.Variable(torch.ones(1), name="cost_weight_aux")),
    )

    input_tensors = {}
    batch_size = 2
    # first check that we can change the current batch size (doubling the size)s
    for var in var_to_cost_functions:
        new_data = torch.ones(batch_size, 1)
        input_tensors[var.name] = new_data
    for aux in aux_to_cost_functions:
        new_data = torch.ones(batch_size, 1)
        input_tensors[aux.name] = new_data
    objective.update(input_tensors=input_tensors)
    assert objective.batch_size == batch_size

    # change one of the variables, no error since batch_size = 1 is broadcastable
    input_tensors["var1"] = torch.ones(1, 1)
    objective.update(input_tensors=input_tensors)
    assert objective.batch_size == batch_size

    # change another variable, this time throws errors since found batch size 2 and 3
    input_tensors["var2"] = torch.ones(batch_size + 1, 1)
    with pytest.raises(ValueError):
        objective.update(input_tensors=input_tensors)

    # change back before testing the aux. variable
    input_tensors["var2"] = torch.ones(batch_size, 1)
    objective.update(input_tensors=input_tensors)  # shouldn't throw error

    # auxiliary variables should also throw error
    input_tensors["aux1"] = torch.ones(batch_size + 1, 1)
    with pytest.raises(ValueError):
        objective.update(input_tensors=input_tensors)


def test_iterator():
    cost_functions, *_ = create_mock_cost_functions()

    objective = th.Objective()
    for f in cost_functions:
        objective.add(f)

    idx = 0
    for f in objective:
        assert f == cost_functions[idx]
        idx += 1


def test_to_dtype():
    objective, *_ = create_objective_with_mock_cost_functions()
    for dtype in [torch.float32, torch.float64]:
        objective.to(dtype=dtype)
        for _, cf in objective.cost_functions.items():
            for var in cf.optim_vars:
                assert var.dtype == dtype
            for aux in cf.aux_vars:
                assert aux.dtype == dtype



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
from collections import OrderedDict
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Union

import torch

from theseus.core.theseus_function import TheseusFunction
from theseus.geometry.manifold import Manifold

from .cost_function import CostFunction
from .cost_weight import CostWeight
from .variable import Variable


# If dtype is None, uses torch.get_default_dtype()
class Objective:
    """An objective function to optimize."""

    def __init__(self, dtype: Optional[torch.dtype] = None):
        # maps variable names to the variable objects
        self.optim_vars: OrderedDict[str, Manifold] = OrderedDict()

        # maps variable names to variables objects, for optimization variables
        # that were registered when adding cost weights.
        self.cost_weight_optim_vars: OrderedDict[str, Manifold] = OrderedDict()

        # maps aux. variable names to the container objects
        self.aux_vars: OrderedDict[str, Variable] = OrderedDict()

        # maps variable name to variable, for any kind of variable added
        self._all_variables: OrderedDict[str, Variable] = OrderedDict()

        # maps cost function names to the cost function objects
        self.cost_functions: OrderedDict[str, CostFunction] = OrderedDict()

        # maps cost weights to the cost functions that use them
        # this is used when deleting cost function to check if the cost weight
        # variables can be deleted as well (when no other function uses them)
        self.cost_functions_for_weights: Dict[CostWeight, List[CostFunction]] = {}

        # ---- The following two methods are used just to get info from
        # ---- the objective, they don't affect the optimization logic.
        # a map from optimization variables to list of theseus functions it's
        # connected to
        self.functions_for_optim_vars: Dict[Manifold, List[TheseusFunction]] = {}

        # a map from all aux. variables to list of theseus functions it's connected to
        self.functions_for_aux_vars: Dict[Variable, List[TheseusFunction]] = {}

        self._batch_size: Optional[int] = None

        self.device: torch.device = torch.device("cpu")

        self.dtype: Optional[torch.dtype] = dtype or torch.get_default_dtype()

        # this increases after every add/erase operation, and it's used to avoid
        # an optimizer to run on a stale version of the objective (since changing the
        # objective structure might break optimizer initialization).
        self.current_version = 0

        # ---- Callbacks for vectorization ---- #
        # This gets replaced when cost function vectorization is used
        self._cost_functions_iterable: Optional[Iterable[CostFunction]] = None

        # Used to vectorize cost functions after update
        self._vectorization_run: Optional[Callable] = None

        # If vectorization is on, this will also handle vectorized containers
        self._vectorization_to: Optional[Callable] = None

        # If vectorization is on, this gets replaced by a vectorized version
        self._retract_method = Objective._retract_base

        # Keeps track of how many variable updates have been made to check
        # if vectorization should be updated
        self._num_updates_variables: Dict[str, int] = {}

        self._last_vectorization_has_grad = False

        self._vectorized = False

    def _add_function_variables(
        self,
        function: TheseusFunction,
        optim_vars: bool = True,
        is_cost_weight: bool = False,
    ):

        if optim_vars:
            function_vars = function.optim_vars
            self_var_to_fn_map = self.functions_for_optim_vars
            self_vars_of_this_type = (
                self.cost_weight_optim_vars if is_cost_weight else self.optim_vars
            )
        else:
            function_vars = function.aux_vars  # type: ignore
            self_var_to_fn_map = self.functions_for_aux_vars  # type: ignore
            self_vars_of_this_type = self.aux_vars  # type: ignore
        for variable in function_vars:
            # Check that variables have name and correct dtype
            if variable.name is None:
                raise ValueError(
                    f"Variables added to an objective must be named, but "
                    f"{function.name} has an unnamed variable."
                )
            if variable.dtype != self.dtype:
                raise ValueError(
                    f"Tried to add variable {variable.name} with dtype "
                    f"{variable.dtype} but objective's dtype is {self.dtype}."
                )
            # Check that names are unique
            if variable.name in self._all_variables:
                if variable is not self._all_variables[variable.name]:
                    raise ValueError(
                        f"Two different variable objects with the "
                        f"same name ({variable.name}) are not allowed "
                        "in the same objective."
                    )
            else:
                self._all_variables[variable.name] = variable
                assert variable not in self_var_to_fn_map
                self_var_to_fn_map[variable] = []

            # add to either self.optim_vars,
            # self.cost_weight_optim_vars or self.aux_vars
            self_vars_of_this_type[variable.name] = variable

            # add to list of functions connected to this variable
            self_var_to_fn_map[variable].append(function)

    # Adds a cost function to the objective
    # Also adds its optimization variables if they haven't been previously added
    # Throws an error if a new variable has the same name of a previously added
    # variable that is not the same object.
    # Does the same for the cost function's auxiliary variables
    # Then does the same with the cost weight's auxiliary variables
    #
    # For now, cost weight "optimization variables" are **NOT** added to the
    # set of objective's variables, they are kept in a separate container.
    # Update method will check if any of these are not registered as
    # cost function variables, and throw a warning.
    def add(self, cost_function: CostFunction):
        # adds the cost function if not already present
        if cost_function.name in self.cost_functions:
            if cost_function is not self.cost_functions[cost_function.name]:
                raise ValueError(
                    f"Two different cost function objects with the "
                    f"same name ({cost_function.name}) are not allowed "
                    "in the same objective."
                )
            else:
                warnings.warn(
                    "This cost function has already been added to the objective, "
                    "nothing to be done."
                )
        else:
            self.cost_functions[cost_function.name] = cost_function

        self.current_version += 1
        # ----- Book-keeping for the cost function ------- #
        # adds information about the optimization variables in this cost function
        self._add_function_variables(cost_function, optim_vars=True)

        # adds information about the auxiliary variables in this cost function
        self._add_function_variables(cost_function, optim_vars=False)

        if cost_function.weight not in self.cost_functions_for_weights:
            # ----- Book-keeping for the cost weight ------- #
            # adds information about the variables in this cost function's weight
            self._add_function_variables(
                cost_function.weight, optim_vars=True, is_cost_weight=True
            )
            # adds information about the auxiliary variables in this cost function's weight
            self._add_function_variables(
                cost_function.weight, optim_vars=False, is_cost_weight=True
            )

            self.cost_functions_for_weights[cost_function.weight] = []

            if cost_function.weight.num_optim_vars() > 0:
                raise RuntimeError(
                    f"The cost weight associated to {cost_function.name} receives one "
                    "or more optimization variables. Differentiating cost "
                    "weights with respect to optimization variables is not currently "
                    "supported, thus jacobians computed by our optimizers will be "
                    "incorrect. You may want to consider moving the weight computation "
                    "inside the cost function, so that the cost weight only receives "
                    "auxiliary variables."
                )

        self.cost_functions_for_weights[cost_function.weight].append(cost_function)

        if self.optim_vars.keys() & self.aux_vars.keys():
            raise ValueError(
                "Objective does not support a variable being both "
                "an optimization variable and an auxiliary variable."
            )

    # returns a reference to the cost function with the given name
    def get_cost_function(self, name: str) -> CostFunction:
        return self.cost_functions.get(name, None)

    # checks if the cost function with the given name is in the objective
    def has_cost_function(self, name: str) -> bool:
        return name in self.cost_functions

    # checks if the optimization variable with the given name is in the objective
    def has_optim_var(self, name: str) -> bool:
        return name in self.optim_vars

    # returns a reference to the optimization variable with the given name
    def get_optim_var(self, name: str) -> Manifold:
        return self.optim_vars.get(name, None)

    # checks if the aux. variable with the given name is in the objective
    def has_aux_var(self, name: str) -> bool:
        return name in self.aux_vars

    # returns a reference to the aux. variable with the given name
    def get_aux_var(self, name: str) -> Variable:
        return self.aux_vars.get(name, None)

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def _erase_function_variables(
        self,
        function: TheseusFunction,
        optim_vars: bool = True,
        is_cost_weight: bool = False,
    ):
        if optim_vars:
            fn_var_list = function.optim_vars
            self_vars_of_this_type = (
                self.cost_weight_optim_vars if is_cost_weight else self.optim_vars
            )
            self_var_to_fn_map = self.functions_for_optim_vars
        else:
            fn_var_list = function.aux_vars  # type: ignore
            self_vars_of_this_type = self.aux_vars  # type: ignore
            self_var_to_fn_map = self.functions_for_aux_vars  # type: ignore

        for variable in fn_var_list:
            cost_fn_idx = self_var_to_fn_map[variable].index(function)
            # remove function from the variable's list of connected cost functions
            del self_var_to_fn_map[variable][cost_fn_idx]
            # if the variable has no other functions, remove it also
            if not self_var_to_fn_map[variable]:
                del self_var_to_fn_map[variable]
                del self_vars_of_this_type[variable.name]

    # Removes a cost function from the objective given its name
    # Also removes any of its variables that are no longer associated to other
    # functions (either cost functions, or cost weights).
    # Does the same for the cost weight, but only if the weight is not associated to
    # any other cost function
    def erase(self, name: str):
        self.current_version += 1
        if name in self.cost_functions:
            cost_function = self.cost_functions[name]
            # erase variables associated to this cost function (if needed)
            self._erase_function_variables(cost_function, optim_vars=True)
            self._erase_function_variables(cost_function, optim_vars=False)

            # delete cost function from list of cost functions connected to its weight
            cost_weight = cost_function.weight
            cost_fn_idx = self.cost_functions_for_weights[cost_weight].index(
                cost_function
            )
            del self.cost_functions_for_weights[cost_weight][cost_fn_idx]

            # No more cost functions associated to this weight, so can also delete
            if len(self.cost_functions_for_weights[cost_weight]) == 0:
                # erase its variables (if needed)
                self._erase_function_variables(
                    cost_weight, optim_vars=True, is_cost_weight=True
                )
                self._erase_function_variables(
                    cost_weight, optim_vars=False, is_cost_weight=True
                )
                del self.cost_functions_for_weights[cost_weight]

            # finally, delete the cost function
            del self.cost_functions[name]
        else:
            warnings.warn(
                "This cost function is not in the objective, nothing to be done."
            )

    # gets the name associated with a cost function object (or None if not present)
    def get_cost_function_name(self, cost_function: CostFunction) -> Optional[str]:
        for name in self.cost_functions:
            if id(cost_function) == id(cost_function):
                return name
        return None

    @staticmethod
    def _get_functions_connected_to_var(
        variable: Union[str, Variable],
        objectives_var_container_dict: "OrderedDict[str, Variable]",
        var_to_cost_fn_map: Dict[Variable, List[TheseusFunction]],
        variable_type: str,
    ) -> List[TheseusFunction]:
        if isinstance(variable, str):
            if variable not in objectives_var_container_dict:
                raise ValueError(
                    f"{variable_type} named {variable} is not in the objective."
                )
            variable = objectives_var_container_dict[variable]
        if variable not in var_to_cost_fn_map:
            raise ValueError(
                f"{variable_type} {variable.name} is not in the objective."
            )
        return var_to_cost_fn_map[variable]

    def get_functions_connected_to_optim_var(
        self, variable: Union[Manifold, str]
    ) -> List[TheseusFunction]:
        return Objective._get_functions_connected_to_var(
            variable,
            self.optim_vars,  # type: ignore
            self.functions_for_optim_vars,  # type: ignore
            "Optimization Variable",
        )

    def get_functions_connected_to_aux_var(
        self, aux_var: Union[Variable, str]
    ) -> List[TheseusFunction]:
        return Objective._get_functions_connected_to_var(
            aux_var, self.aux_vars, self.functions_for_aux_vars, "Auxiliary Variable"
        )

    # sum of cost function dimensions
    def dim(self) -> int:
        err_dim = 0
        for cost_function in self.cost_functions.values():
            err_dim += cost_function.dim()
        return err_dim

    # number of (cost functions, variables)
    def size(self) -> tuple:
        return len(self.cost_functions), len(self.optim_vars)

    # number of cost functions
    def size_cost_functions(self) -> int:
        return len(self.cost_functions)

    # number of variables
    def size_variables(self) -> int:
        return len(self.optim_vars)

    # number of auxiliary variables
    def size_aux_vars(self) -> int:
        return len(self.aux_vars)

    def error(
        self,
        input_tensors: Optional[Dict[str, torch.Tensor]] = None,
        also_update: bool = False,
    ) -> torch.Tensor:
        old_tensors = {}
        if input_tensors is not None:
            if not also_update:
                for var in self.optim_vars:
                    old_tensors[var] = self.optim_vars[var].tensor
            self.update(input_tensors=input_tensors)

        error_vector = torch.cat(
            [cf.weighted_error() for cf in self._get_iterator()], dim=1
        )

        if input_tensors is not None and not also_update:
            self.update(old_tensors)
        return error_vector

    def error_squared_norm(
        self,
        input_tensors: Optional[Dict[str, torch.Tensor]] = None,
        also_update: bool = False,
    ) -> torch.Tensor:
        return (
            self.error(input_tensors=input_tensors, also_update=also_update) ** 2
        ).sum(dim=1)

    def copy(self) -> "Objective":
        new_objective = Objective(dtype=self.dtype)

        # First copy all individual cost weights
        old_to_new_cost_weight_map: Dict[CostWeight, CostWeight] = {}
        for cost_weight in self.cost_functions_for_weights:
            new_cost_weight = cost_weight.copy(
                new_name=cost_weight.name, keep_variable_names=True
            )
            old_to_new_cost_weight_map[cost_weight] = new_cost_weight

        # Now copy the cost functions and assign the corresponding cost weight copy
        new_cost_functions: List[CostFunction] = []
        for cost_function in self.cost_functions.values():
            new_cost_function = cost_function.copy(
                new_name=cost_function.name, keep_variable_names=True
            )
            # we assign the allocated weight copies to avoid saving duplicates
            new_cost_function.weight = old_to_new_cost_weight_map[cost_function.weight]
            new_cost_functions.append(new_cost_function)

        # Handle case where a variable is copied in 2+ cost functions or cost weights,
        # since only a single copy should be maintained by objective
        for cost_function in new_cost_functions:
            # CostFunction
            for i, var in enumerate(cost_function.optim_vars):
                if new_objective.has_optim_var(var.name):
                    cost_function.set_optim_var_at(
                        i, new_objective.optim_vars[var.name]
                    )
            for i, aux_var in enumerate(cost_function.aux_vars):
                if new_objective.has_aux_var(aux_var.name):
                    cost_function.set_aux_var_at(
                        i, new_objective.aux_vars[aux_var.name]
                    )
            # CostWeight
            for i, var in enumerate(cost_function.weight.optim_vars):
                if var.name in new_objective.cost_weight_optim_vars:
                    cost_function.weight.set_optim_var_at(
                        i, new_objective.cost_weight_optim_vars[var.name]
                    )
            for i, aux_var in enumerate(cost_function.weight.aux_vars):
                if new_objective.has_aux_var(aux_var.name):
                    cost_function.weight.set_aux_var_at(
                        i, new_objective.aux_vars[aux_var.name]
                    )
            new_objective.add(cost_function)
        return new_objective

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]
        the_copy = self.copy()
        memo[id(self)] = the_copy
        return the_copy

    def update(self, input_tensors: Optional[Dict[str, torch.Tensor]] = None):
        self._batch_size = None

        def _get_batch_size(batch_sizes: Sequence[int]) -> int:
            unique_batch_sizes = set(batch_sizes)
            if len(unique_batch_sizes) == 1:
                return batch_sizes[0]
            if len(unique_batch_sizes) == 2:
                min_bs = min(unique_batch_sizes)
                max_bs = max(unique_batch_sizes)
                if min_bs == 1:
                    return max_bs
            raise ValueError("Provided tensors must be broadcastable.")

        input_tensors = input_tensors or {}
        for var_name, tensor in input_tensors.items():
            if tensor.ndim < 2:
                raise ValueError(
                    f"Input tensors must have a batch dimension and "
                    f"one ore more data dimensions, but tensor.ndim={tensor.ndim} for "
                    f"tensor with name {var_name}."
                )
            if var_name in self.optim_vars:
                self.optim_vars[var_name].update(tensor)
            elif var_name in self.aux_vars:
                self.aux_vars[var_name].update(tensor)
            elif var_name in self.cost_weight_optim_vars:
                self.cost_weight_optim_vars[var_name].update(tensor)
                warnings.warn(
                    "Updated a variable declared as optimization, but it is "
                    "only associated to cost weights and not to any cost functions. "
                    "Theseus optimizers will only update optimization variables "
                    "that are associated to one or more cost functions."
                )
            else:
                warnings.warn(
                    f"Attempted to update a tensor with name {var_name}, "
                    "which is not associated to any variable in the objective."
                )

        # Check that the batch size of all tensors is consistent after update
        batch_sizes = [v.tensor.shape[0] for v in self.optim_vars.values()]
        batch_sizes.extend([v.tensor.shape[0] for v in self.aux_vars.values()])
        self._batch_size = _get_batch_size(batch_sizes)

    def _vectorization_needs_update(self):
        num_updates = dict(
            (name, v._num_updates) for name, v in self._all_variables.items()
        )
        needs = False
        if num_updates != self._num_updates_variables:
            self._num_updates_variables = num_updates
            needs = True

        if torch.is_grad_enabled():
            if not self._last_vectorization_has_grad:
                needs = True
        return needs

    def update_vectorization_if_needed(self):
        if self.vectorized and self._vectorization_needs_update():
            if self._batch_size is None:
                self.update()
            self._vectorization_run()
            self._last_vectorization_has_grad = torch.is_grad_enabled()

    # iterates over cost functions
    def __iter__(self):
        return iter([cf for cf in self.cost_functions.values()])

    def _get_iterator(self):
        self.update_vectorization_if_needed()
        if self._cost_functions_iterable is None:
            return iter([cf for cf in self.cost_functions.values()])
        return iter([cf for cf in self._cost_functions_iterable])

    # Applies to() with given args to all tensors in the objective
    def to(self, *args, **kwargs):
        for cost_function in self.cost_functions.values():
            cost_function.to(*args, **kwargs)
        device, dtype, *_ = torch._C._nn._parse_to(*args, **kwargs)
        self.device = device or self.device
        self.dtype = dtype or self.dtype
        if self._vectorization_to is not None:
            self._vectorization_to(*args, **kwargs)

    @staticmethod
    def _retract_base(
        delta: torch.Tensor,
        ordering: Iterable[Manifold],
        ignore_mask: Optional[torch.Tensor] = None,
        force_update: bool = False,
    ):
        var_idx = 0
        for var in ordering:
            new_var = var.retract(delta[:, var_idx : var_idx + var.dof()])
            if ignore_mask is None or force_update:
                var.update(new_var.tensor)
            else:
                var.update(new_var.tensor, batch_ignore_mask=ignore_mask)
            var_idx += var.dof()

    def retract_optim_vars(
        self,
        delta: torch.Tensor,
        ordering: Iterable[Manifold],
        ignore_mask: Optional[torch.Tensor] = None,
        force_update: bool = False,
    ):
        self._retract_method(
            delta, ordering, ignore_mask=ignore_mask, force_update=force_update
        )

    def _enable_vectorization(
        self,
        cost_fns_iter: Iterable[CostFunction],
        vectorization_run_fn: Callable,
        vectorized_to: Callable,
        vectorized_retract_fn: Callable,
        enabler: Any,
    ):
        # Hacky way to make Vectorize a "friend" class
        assert (
            enabler.__module__ == "theseus.core.vectorizer"
            and enabler.__class__.__name__ == "Vectorize"
        )
        self._cost_functions_iterable = cost_fns_iter
        self._vectorization_run = vectorization_run_fn
        self._vectorization_to = vectorized_to
        self._retract_method = vectorized_retract_fn
        self._vectorized = True

    # Making public, since this should be a safe operation
    def disable_vectorization(self):
        self._cost_functions_iterable = None
        self._vectorization_run = None
        self._vectorization_to = None
        self._retract_method = Objective._retract_base
        self._vectorized = False

    @property
    def vectorized(self):
        assert (
            (not self._vectorized)
            == (self._cost_functions_iterable is None)
            == (self._vectorization_run is None)
            == (self._vectorization_to is None)
            == (self._retract_method is Objective._retract_base)
        )
        return self._vectorized



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th
from theseus.constants import TEST_EPS
from theseus.core.tests.common import check_copy_var
from theseus.utils import numeric_jacobian

from .common import (
    check_adjoint,
    check_compose,
    check_exp_map,
    check_inverse,
    check_log_map,
    check_normalize,
    check_projection_for_compose,
    check_projection_for_exp_map,
    check_projection_for_inverse,
    check_projection_for_rotate_and_transform,
)


def create_random_se2(batch_size, rng, dtype=torch.float64):
    theta = torch.rand(batch_size, 1, generator=rng) * 2 * np.pi - np.pi
    u = torch.randn(batch_size, 2)
    tangent_vector = torch.cat([u, theta], dim=1)
    return th.SE2.exp_map(tangent_vector.to(dtype=dtype))


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_exp_map(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        theta = torch.from_numpy(np.linspace(-np.pi, np.pi, batch_size))
        u = torch.randn(batch_size, 2, dtype=dtype, generator=rng)
        tangent_vector = torch.cat([u, theta.unsqueeze(1)], dim=1)
        check_exp_map(tangent_vector.to(dtype=dtype), th.SE2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_log_map(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        theta = torch.from_numpy(np.linspace(-np.pi, np.pi, batch_size))
        u = torch.randn(batch_size, 2, dtype=dtype, generator=rng)
        tangent_vector = torch.cat([u, theta.unsqueeze(1)], dim=1)
        check_log_map(tangent_vector, th.SE2)
        check_projection_for_exp_map(tangent_vector, th.SE2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_compose(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        se2_1 = th.SE2.rand(batch_size, generator=rng, dtype=dtype)
        se2_2 = th.SE2.rand(batch_size, generator=rng, dtype=dtype)
        check_compose(se2_1, se2_2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_inverse(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        se2 = th.SE2.rand(batch_size, generator=rng, dtype=dtype)
        check_inverse(se2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_adjoint(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        se2 = th.SE2.rand(batch_size, generator=rng, dtype=dtype)
        tangent = torch.randn(batch_size, 3, dtype=dtype)
        check_adjoint(se2, tangent)


def test_copy():
    rng = torch.Generator()
    se2 = th.SE2.rand(1, generator=rng, dtype=torch.float64)
    check_copy_var(se2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_transform_from_and_to(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for _ in range(10):  # repeat a few times
        for batch_size_se2 in [1, 20, 100]:
            for batch_size_pnt in [1, 20, 100]:
                if (
                    batch_size_se2 != 1
                    and batch_size_pnt != 1
                    and batch_size_pnt != batch_size_se2
                ):
                    continue

                se2 = th.SE2.rand(batch_size_se2, generator=rng, dtype=dtype)
                point_tensor = torch.randn(batch_size_pnt, 2, dtype=dtype)
                point_tensor_ext = torch.cat(
                    [point_tensor, torch.ones(batch_size_pnt, 1, dtype=dtype)], dim=1
                )

                jacobians_to = []
                point_to = se2.transform_to(point_tensor, jacobians=jacobians_to)
                expected_to = (
                    se2.inverse().to_matrix().double()
                    @ point_tensor_ext.unsqueeze(2).double()
                )[:, :2]
                jacobians_from = []
                point_from = se2.transform_from(point_to, jacobians_from)

                # Check the operation result
                assert torch.allclose(
                    expected_to.squeeze(2), point_to.tensor.double(), atol=TEST_EPS
                )
                assert torch.allclose(
                    point_tensor.double(), point_from.tensor.double(), atol=TEST_EPS
                )

                # Check the jacobians
                se2_double = se2.copy()
                se2_double.to(torch.float64)
                expected_jac = numeric_jacobian(
                    lambda groups: groups[0].transform_to(groups[1]),
                    [se2_double, th.Point2(point_tensor.double())],
                    function_dim=2,
                )

                assert jacobians_to[0].shape == expected_jac[0].shape
                assert jacobians_to[1].shape == expected_jac[1].shape
                assert torch.allclose(
                    jacobians_to[0].double(), expected_jac[0], atol=TEST_EPS
                )
                assert torch.allclose(
                    jacobians_to[1].double(), expected_jac[1], atol=TEST_EPS
                )

                expected_jac = numeric_jacobian(
                    lambda groups: groups[0].transform_from(groups[1]),
                    [se2_double, th.Point2(point_to.tensor.double())],
                    function_dim=2,
                )
                assert jacobians_from[0].shape == expected_jac[0].shape
                assert jacobians_from[1].shape == expected_jac[1].shape
                assert torch.allclose(
                    jacobians_from[0].double(), expected_jac[0], atol=TEST_EPS
                )
                assert torch.allclose(
                    jacobians_from[1].double(), expected_jac[1], atol=TEST_EPS
                )


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_xy_jacobian(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        se2 = th.SE2.rand(batch_size, generator=rng, dtype=dtype)
        jacobian = []
        se2.xy(jacobians=jacobian)
        expected_jac = numeric_jacobian(
            lambda groups: th.Point2(groups[0].xy()), [se2], function_dim=2
        )
        torch.allclose(jacobian[0], expected_jac[0])


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_theta_jacobian(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        se2 = th.SE2.rand(batch_size, generator=rng, dtype=dtype)
        jacobian = []
        se2.theta(jacobians=jacobian)
        expected_jac = numeric_jacobian(
            lambda groups: th.Vector(tensor=groups[0].theta()), [se2], function_dim=1
        )
        torch.allclose(jacobian[0], expected_jac[0])


def test_projection():
    rng = torch.Generator()
    rng.manual_seed(0)
    for _ in range(10):  # repeat a few times
        for batch_size in [1, 20, 100]:
            # Test SE2.transform_to
            check_projection_for_rotate_and_transform(
                th.SE2, th.Point2, th.SE2.transform_to, batch_size, rng
            )

            # Test SE2.transform_from
            check_projection_for_rotate_and_transform(
                th.SE2, th.Point2, th.SE2.transform_from, batch_size, rng
            )

            # Test SE2.compose
            check_projection_for_compose(th.SE2, batch_size, rng)

            # Test SE2.inverse
            check_projection_for_inverse(th.SE2, batch_size, rng)


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_normalization(batch_size, dtype):
    check_normalize(th.SE2, batch_size, dtype)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th
from theseus.constants import TEST_EPS
from theseus.utils import numeric_jacobian

from .common import (
    check_adjoint,
    check_compose,
    check_exp_map,
    check_jacobian_for_local,
    check_projection_for_compose,
    check_projection_for_exp_map,
    check_projection_for_inverse,
    check_projection_for_log_map,
    check_projection_for_rotate_and_transform,
    check_so3_se3_normalize,
)


def check_SO3_log_map(tangent_vector, atol=1e-7):
    error = (tangent_vector - th.SO3.exp_map(tangent_vector).log_map()).norm(dim=1)
    error = torch.minimum(error, (error - 2 * np.pi).abs())
    assert torch.allclose(error, torch.zeros_like(error), atol=atol)


def check_SO3_to_quaternion(so3: th.SO3, atol=1e-10):
    quaternions = so3.to_quaternion()
    assert torch.allclose(
        th.SO3(quaternion=quaternions).to_matrix(), so3.to_matrix(), atol=atol
    )


def _create_tangent_vector(batch_size, ang_factor, rng, dtype):
    tangent_vector = torch.rand(batch_size, 3, generator=rng, dtype=dtype) - 0.5
    tangent_vector /= tangent_vector.norm(dim=1, keepdim=True)
    tangent_vector *= ang_factor
    return tangent_vector


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
@pytest.mark.parametrize(
    "ang_factor", [None, 1e-5, 3e-3, 2 * np.pi - 1e-11, np.pi - 1e-7, np.pi - 1e-11]
)
def test_exp_map(batch_size, dtype, ang_factor):
    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 2e-4 if dtype == torch.float32 else 1e-6

    if ang_factor is None:
        ang_factor = (
            torch.rand(batch_size, 1, generator=rng, dtype=dtype) * 2 * np.pi - np.pi
        )

    tangent_vector = _create_tangent_vector(batch_size, ang_factor, rng, dtype)
    check_exp_map(tangent_vector, th.SO3, atol=ATOL)
    check_projection_for_exp_map(tangent_vector, th.SO3, atol=ATOL)


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
@pytest.mark.parametrize(
    "ang_factor", [None, 1e-5, 3e-3, 2 * np.pi - 1e-11, np.pi - 1e-11]
)
def test_log_map(batch_size, dtype, ang_factor):
    if dtype == torch.float32 and ang_factor == np.pi - 1e-11:
        return

    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 1e-3 if dtype == torch.float32 else 1e-8
    PROJECTION_ATOL = 1e-2 if dtype == torch.float32 else 1e-8

    if ang_factor is None:
        ang_factor = (
            torch.rand(batch_size, 1, generator=rng, dtype=dtype) * 2 * np.pi - np.pi
        )

    tangent_vector = _create_tangent_vector(batch_size, ang_factor, rng, dtype)
    check_SO3_log_map(tangent_vector, atol=ATOL)
    check_projection_for_log_map(tangent_vector, th.SO3, atol=PROJECTION_ATOL)


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
@pytest.mark.parametrize(
    "ang_factor", [None, 1e-5, 3e-3, 2 * np.pi - 1e-11, np.pi - 1e-11]
)
def test_quaternion(batch_size, dtype, ang_factor):
    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 1e-3 if dtype == torch.float32 else 1e-8

    if ang_factor is None:
        ang_factor = (
            torch.rand(batch_size, 1, generator=rng, dtype=dtype) * 2 * np.pi - np.pi
        )

    tangent_vector = _create_tangent_vector(batch_size, ang_factor, rng, dtype)
    so3 = th.SO3.exp_map(tangent_vector)
    check_SO3_to_quaternion(so3, atol=ATOL)


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_adjoint(batch_size, dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    so3 = th.SO3.rand(batch_size, generator=rng, dtype=dtype)
    tangent = torch.randn(batch_size, 3, dtype=dtype)
    check_adjoint(so3, tangent)


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_compose(batch_size, dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    so3_1 = th.SO3.rand(batch_size, generator=rng, dtype=dtype)
    so3_2 = th.SO3.rand(batch_size, generator=rng, dtype=dtype)
    check_compose(so3_1, so3_2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_rotate_and_unrotate(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for _ in range(10):  # repeat a few times
        for batch_size_group in [1, 20, 100]:
            for batch_size_pnt in [1, 20, 100]:
                if (
                    batch_size_group != 1
                    and batch_size_pnt != 1
                    and batch_size_pnt != batch_size_group
                ):
                    continue

                so3 = th.SO3.rand(batch_size_group, generator=rng, dtype=dtype)
                point_tensor = torch.randn(
                    batch_size_pnt, 3, generator=rng, dtype=dtype
                )

                jacobians_rotate = []
                rotated_point = so3.rotate(point_tensor, jacobians=jacobians_rotate)
                expected_rotated_data = so3.to_matrix() @ point_tensor.unsqueeze(2)
                jacobians_unrotate = []
                unrotated_point = so3.unrotate(rotated_point, jacobians_unrotate)

                # Check the operation result
                assert torch.allclose(
                    expected_rotated_data.squeeze(2),
                    rotated_point.tensor,
                    atol=TEST_EPS,
                )
                assert torch.allclose(
                    point_tensor, unrotated_point.tensor, atol=TEST_EPS
                )

                # Check the jacobians
                # function_dim = 3 because rotate(so3, (x, y, z)) --> (x_new, y_new, z_new)
                so3_double = so3.copy()
                so3_double.to(torch.float64)
                expected_jac = numeric_jacobian(
                    lambda groups: groups[0].rotate(groups[1]),
                    [so3_double, th.Point3(point_tensor.double())],
                    function_dim=3,
                )
                assert torch.allclose(
                    jacobians_rotate[0].double(), expected_jac[0], atol=TEST_EPS
                )
                assert torch.allclose(
                    jacobians_rotate[1].double(), expected_jac[1], atol=TEST_EPS
                )
                expected_jac = numeric_jacobian(
                    lambda groups: groups[0].unrotate(groups[1]),
                    [so3_double, th.Vector(tensor=rotated_point.tensor.double())],
                    delta_mag=1e-5,
                    function_dim=3,
                )
                assert torch.allclose(
                    jacobians_unrotate[0].double(), expected_jac[0], atol=TEST_EPS
                )
                assert torch.allclose(
                    jacobians_unrotate[1].double(), expected_jac[1], atol=TEST_EPS
                )


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_projection(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for _ in range(10):  # repeat a few times
        for batch_size in [1, 20, 100]:
            # Test SO3.rotate
            check_projection_for_rotate_and_transform(
                th.SO3, th.Point3, th.SO3.rotate, batch_size, rng, dtype=dtype
            )

            # Test SO3.unrotate
            check_projection_for_rotate_and_transform(
                th.SO3, th.Point3, th.SO3.unrotate, batch_size, rng, dtype=dtype
            )

            # Test SO3.compose
            check_projection_for_compose(th.SO3, batch_size, rng, dtype=dtype)

            # Test SO3.inverse
            check_projection_for_inverse(th.SO3, batch_size, rng, dtype=dtype)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_local_map(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 3e-5 if dtype == torch.float32 else 1e-7

    for batch_size in [1, 20, 100]:
        group0 = th.SO3.rand(batch_size, dtype=dtype)
        group1 = th.SO3.rand(batch_size, dtype=dtype)

        check_jacobian_for_local(
            group0, group1, Group=th.SO3, is_projected=True, atol=ATOL
        )


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_normalization(batch_size, dtype):
    check_so3_se3_normalize(th.SO3, batch_size, dtype)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th
from theseus.constants import TEST_EPS
from theseus.utils import numeric_jacobian

from .common import (
    check_adjoint,
    check_compose,
    check_exp_map,
    check_inverse,
    check_jacobian_for_local,
    check_projection_for_compose,
    check_projection_for_exp_map,
    check_projection_for_inverse,
    check_projection_for_log_map,
    check_projection_for_rotate_and_transform,
    check_so3_se3_normalize,
)


def check_SE3_log_map(tangent_vector, atol=TEST_EPS):
    g = th.SE3.exp_map(tangent_vector)
    assert torch.allclose(th.SE3.exp_map(g.log_map()).tensor, g.tensor, atol=atol)


def _create_tangent_vector(batch_size, ang_factor, rng, dtype):
    tangent_vector_ang = torch.rand(batch_size, 3, generator=rng, dtype=dtype) - 0.5
    tangent_vector_ang /= tangent_vector_ang.norm(dim=1, keepdim=True)
    tangent_vector_ang *= ang_factor
    tangent_vector_lin = torch.randn(batch_size, 3, generator=rng, dtype=dtype)
    tangent_vector = torch.cat([tangent_vector_lin, tangent_vector_ang], dim=1)
    return tangent_vector


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
@pytest.mark.parametrize(
    "ang_factor", [None, 1e-5, 3e-3, 2 * np.pi - 1e-11, np.pi - 1e-11]
)
def test_exp_map(batch_size, dtype, ang_factor):
    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 2e-4 if dtype == torch.float32 else 1e-6

    if ang_factor is None:
        ang_factor = (
            torch.rand(batch_size, 1, generator=rng, dtype=dtype) * 2 * np.pi - np.pi
        )
    tangent_vector = _create_tangent_vector(batch_size, ang_factor, rng, dtype)
    check_exp_map(tangent_vector, th.SE3, atol=ATOL)
    check_projection_for_exp_map(tangent_vector, th.SE3, atol=ATOL)


# This test checks that cross products are done correctly
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
@pytest.mark.parametrize(
    "ang_factor", [None, 1e-5, 3e-3, 2 * np.pi - 1e-11, np.pi - 1e-11]
)
def test_batch_size_3_exp_map(dtype, ang_factor):
    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 1e-4 if dtype == torch.float32 else 1e-6

    if ang_factor is None:
        ang_factor = torch.rand(6, 1, generator=rng, dtype=dtype) * 2 * np.pi - np.pi
    tangent_vector = _create_tangent_vector(6, ang_factor, rng, dtype)

    jac, jac1, jac2 = [], [], []
    g = th.SE3.exp_map(tangent_vector, jac)
    g1 = th.SE3.exp_map(tangent_vector[:3], jac1)
    g2 = th.SE3.exp_map(tangent_vector[3:], jac2)

    torch.allclose(g.tensor[:3], g1.tensor, atol=1e-6)
    torch.allclose(g.tensor[3:], g2.tensor, atol=1e-6)
    torch.allclose(jac[0][:3], jac1[0], atol=ATOL)
    torch.allclose(jac[0][3:], jac2[0], atol=ATOL)


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
@pytest.mark.parametrize(
    "ang_factor", [None, 1e-5, 3e-3, 2 * np.pi - 1e-11, np.pi - 1e-7, np.pi - 1e-10]
)
def test_log_map(batch_size, dtype, ang_factor):
    if dtype == torch.float32 and ang_factor == np.pi - 1e-10:
        return

    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 1e-3 if dtype == torch.float32 else 1e-8
    PROJECTION_ATOL = 1e-2 if dtype == torch.float32 else 1e-8
    if ang_factor is None:
        ang_factor = (
            torch.rand(batch_size, 1, generator=rng, dtype=dtype) * 2 * np.pi - np.pi
        )
    tangent_vector = _create_tangent_vector(batch_size, ang_factor, rng, dtype)
    check_SE3_log_map(tangent_vector, atol=ATOL)
    check_projection_for_log_map(tangent_vector, th.SE3, atol=PROJECTION_ATOL)


# This test checks that cross products are done correctly
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
@pytest.mark.parametrize(
    "ang_factor", [None, 1e-5, 3e-3, 2 * np.pi - 1e-11, np.pi - 1e-11]
)
def test_batch_size_3_log_map(dtype, ang_factor):
    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 1e-3 if dtype == torch.float32 else 1e-6

    if ang_factor is None:
        ang_factor = torch.rand(6, 1, generator=rng, dtype=dtype) * 2 * np.pi - np.pi
    tangent_vector = _create_tangent_vector(6, ang_factor, rng, dtype)

    g = th.SE3.exp_map(tangent_vector)
    g1 = th.SE3(tensor=g.tensor[:3])
    g2 = th.SE3(tensor=g.tensor[3:])

    jac, jac1, jac2 = [], [], []
    d = g.log_map(jac)
    d1 = g1.log_map(jac1)
    d2 = g2.log_map(jac2)

    torch.allclose(d[:3], d1, atol=ATOL)
    torch.allclose(d[3:], d2, atol=ATOL)
    torch.allclose(jac[0][:3], jac1[0], atol=ATOL)
    torch.allclose(jac[0][3:], jac2[0], atol=ATOL)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_compose(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        se3_1 = th.SE3.rand(batch_size, generator=rng, dtype=dtype)
        se3_2 = th.SE3.rand(batch_size, generator=rng, dtype=dtype)
        check_compose(se3_1, se3_2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_inverse(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        se3 = th.SE3.rand(batch_size, generator=rng, dtype=dtype)
        check_inverse(se3)


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_adjoint(batch_size, dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    se3 = th.SE3.rand(batch_size, generator=rng, dtype=dtype)
    tangent = torch.randn(batch_size, 6, generator=rng, dtype=dtype)
    check_adjoint(se3, tangent)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_transform_from_and_to(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for _ in range(10):  # repeat a few times
        for batch_size_group in [1, 20]:
            for batch_size_pnt in [1, 20]:
                if (
                    batch_size_group != 1
                    and batch_size_pnt != 1
                    and batch_size_pnt != batch_size_group
                ):
                    continue
                se3 = th.SE3.rand(batch_size_group, generator=rng, dtype=dtype)
                point_tensor = torch.randn(
                    batch_size_pnt, 3, generator=rng, dtype=dtype
                )
                point_tensor_ext = torch.cat(
                    (point_tensor, torch.ones(batch_size_pnt, 1, dtype=dtype)), dim=1
                )

                jacobians_to = []
                point_to = se3.transform_to(point_tensor, jacobians=jacobians_to)
                expected_to = (
                    se3.inverse().to_matrix().double()
                    @ point_tensor_ext.double().unsqueeze(2)
                )[:, :3].squeeze(2)
                jacobians_from = []
                point_from = se3.transform_from(point_to, jacobians_from)

                # Check the operation result
                assert torch.allclose(
                    expected_to, point_to.tensor.double(), atol=TEST_EPS
                )
                assert torch.allclose(point_tensor, point_from.tensor, atol=5e-7)

                # Check the jacobians
                se3_double = se3.copy()
                se3_double.to(torch.float64)
                expected_jac = numeric_jacobian(
                    lambda groups: groups[0].transform_to(groups[1]),
                    [se3_double, th.Point3(point_tensor.double())],
                    function_dim=3,
                )
                assert torch.allclose(
                    jacobians_to[0].double(), expected_jac[0], atol=5e-7
                )
                assert torch.allclose(
                    jacobians_to[1].double(), expected_jac[1], atol=TEST_EPS
                )
                expected_jac = numeric_jacobian(
                    lambda groups: groups[0].transform_from(groups[1]),
                    [se3_double, th.Vector(tensor=point_to.tensor.double())],
                    delta_mag=1e-5,
                    function_dim=3,
                )
                assert torch.allclose(
                    jacobians_from[0].double(), expected_jac[0], atol=TEST_EPS
                )
                assert torch.allclose(
                    jacobians_from[1].double(), expected_jac[1], atol=TEST_EPS
                )


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_projection(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    for _ in range(10):  # repeat a few times
        for batch_size in [1, 20]:
            # Test SE3.transform_to
            check_projection_for_rotate_and_transform(
                th.SE3, th.Point3, th.SE3.transform_to, batch_size, rng, dtype=dtype
            )

            # Test SE3.transform_from
            check_projection_for_rotate_and_transform(
                th.SE3, th.Point3, th.SE3.transform_from, batch_size, rng, dtype=dtype
            )

            # Test SE3.compose
            check_projection_for_compose(th.SE3, batch_size, rng, dtype=dtype)

            # Test SE3.inverse
            check_projection_for_inverse(th.SE3, batch_size, rng, dtype=dtype)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_local_map(dtype):
    rng = torch.Generator()
    rng.manual_seed(0)
    ATOL = 3e-5 if dtype == torch.float32 else 1e-7

    for batch_size in [1, 20, 100]:
        group0 = th.SE3.rand(batch_size, dtype=dtype)
        group1 = th.SE3.rand(batch_size, dtype=dtype)
        check_jacobian_for_local(
            group0, group1, Group=th.SE3, is_projected=True, atol=ATOL
        )


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_normalization(batch_size, dtype):
    check_so3_se3_normalize(th.SE3, batch_size, dtype)
